from unittest import mock

import pytest
from google.protobuf import wrappers_pb2
from google.protobuf.any_pb2 import Any as ProtoAny
from google.protobuf.message import Message as ProtoMessage
from google.protobuf.struct_pb2 import Struct

from lekko_client.clients.config_client import AnyProto
from lekko_client.clients.distribution_client import CachedDistributionClient
from lekko_client.evaluation.evaluation import EvaluationResult
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    RepositoryKey,
)
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import Feature
from lekko_client.helpers import convert_context, get_context_keys
from lekko_client.models import FeatureData


@pytest.mark.parametrize(
    "fn_under_test,expected",
    [
        ("get_bool", wrappers_pb2.BoolValue(value=True)),
        ("get_int", wrappers_pb2.Int64Value(value=10)),
        ("get_float", wrappers_pb2.FloatValue(value=2.5)),
        ("get_string", wrappers_pb2.StringValue(value="test_value")),
    ],
)
def test_scalar(
    mock_distribution_client: CachedDistributionClient,
    test_feature_no_constraints: Feature,
    fn_under_test: str,
    expected: ProtoMessage,
):
    mock_distribution_client.store.get.return_value = FeatureData("test_sha", test_feature_no_constraints)

    expected_any = ProtoAny()
    expected_any.Pack(expected)
    namespace = "test_namespace"
    key = test_feature_no_constraints.key
    ctx = {"ctx_key": "ctx_val"}
    with mock.patch(
        "lekko_client.clients.distribution_client.evaluate", return_value=EvaluationResult(value=expected_any, path=[1])
    ) as mock_eval:
        getattr(mock_distribution_client, fn_under_test)(namespace, key, ctx)
        res_unpacked = type(expected)()
        assert expected_any.Unpack(res_unpacked)
        assert res_unpacked == expected

        mock_eval.assert_called_once_with(test_feature_no_constraints, namespace, convert_context(ctx))
        mock_distribution_client.store.get.assert_called_once_with(namespace, key)

        # Not worth freezing time to use assert_called_once_with
        (flag_evaluation,) = mock_distribution_client.events_batcher.add_event.call_args_list[0].args
        assert flag_evaluation.repo_key == RepositoryKey(owner_name="owner", repo_name="repo")
        assert flag_evaluation.commit_sha == "test_commit_sha"
        assert flag_evaluation.feature_sha == "test_sha"
        assert flag_evaluation.namespace_name == namespace
        assert flag_evaluation.feature_name == key
        assert flag_evaluation.context_keys == get_context_keys(convert_context(ctx))
        assert flag_evaluation.result_path == [1]


@pytest.mark.parametrize(
    "expected",
    [
        [1, 2, 3],
        {"test": "dict"},
        "str",
        1.2,
        True,
    ],
)
def test_get_json(mock_distribution_client, test_feature_no_constraints, expected):
    s = Struct()
    s.update({"key": expected})
    expected_value = s.fields["key"]
    any_proto = AnyProto()
    any_proto.Pack(expected_value)

    mock_distribution_client.store.get.return_value = FeatureData("test_sha", test_feature_no_constraints)

    with mock.patch(
        "lekko_client.clients.distribution_client.evaluate", return_value=EvaluationResult(value=any_proto, path=[1])
    ):
        assert expected == mock_distribution_client.get_json("namespace", "key", {})


def test_get_proto_by_type(mock_distribution_client, test_feature_no_constraints):
    any_proto = AnyProto()
    int_proto = wrappers_pb2.Int32Value(value=10)
    any_proto.Pack(int_proto)
    mock_distribution_client.store.get.return_value = FeatureData("test_sha", test_feature_no_constraints)

    with mock.patch(
        "lekko_client.clients.distribution_client.evaluate", return_value=EvaluationResult(value=any_proto, path=[1])
    ):
        assert int_proto == mock_distribution_client.get_proto_by_type("namespace", "key", {}, wrappers_pb2.Int32Value)


def test_get_proto(mock_distribution_client, test_feature_no_constraints):
    any_proto = AnyProto()
    int_proto = wrappers_pb2.Int32Value(value=10)
    any_proto.Pack(int_proto)
    mock_distribution_client.store.get.return_value = FeatureData("test_sha", test_feature_no_constraints)

    with mock.patch(
        "lekko_client.clients.distribution_client.evaluate", return_value=EvaluationResult(value=any_proto, path=[1])
    ):
        assert int_proto == mock_distribution_client.get_proto("namespace", "key", {})

        with mock.patch("google.protobuf.symbol_database.SymbolDatabase.GetSymbol", side_effect=KeyError):
            assert any_proto == mock_distribution_client.get_proto("namespace", "key", {})



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 Lekko Technologies, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import lekko_client.gen.lekko.rules.v1beta2.rules_pb2
import lekko_client.gen.lekko.rules.v1beta3.rules_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _FeatureType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _FeatureTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_FeatureType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    FEATURE_TYPE_UNSPECIFIED: _FeatureType.ValueType  # 0
    FEATURE_TYPE_BOOL: _FeatureType.ValueType  # 1
    FEATURE_TYPE_INT: _FeatureType.ValueType  # 2
    FEATURE_TYPE_FLOAT: _FeatureType.ValueType  # 3
    FEATURE_TYPE_STRING: _FeatureType.ValueType  # 4
    FEATURE_TYPE_JSON: _FeatureType.ValueType  # 5
    FEATURE_TYPE_PROTO: _FeatureType.ValueType  # 6

class FeatureType(_FeatureType, metaclass=_FeatureTypeEnumTypeWrapper):
    """Enumerates the canonical types that lekko supports"""

FEATURE_TYPE_UNSPECIFIED: FeatureType.ValueType  # 0
FEATURE_TYPE_BOOL: FeatureType.ValueType  # 1
FEATURE_TYPE_INT: FeatureType.ValueType  # 2
FEATURE_TYPE_FLOAT: FeatureType.ValueType  # 3
FEATURE_TYPE_STRING: FeatureType.ValueType  # 4
FEATURE_TYPE_JSON: FeatureType.ValueType  # 5
FEATURE_TYPE_PROTO: FeatureType.ValueType  # 6
global___FeatureType = FeatureType

@typing_extensions.final
class Feature(google.protobuf.message.Message):
    """A prototype of the wrapper type that will be used to house all feature flags for
    the 'homegrown' feature flagging solution:
    User-defined proto defintions, and a tree-based constraints system.
    A real-life example of this in practice is visualized here:
    https://lucid.app/lucidchart/f735298f-db2c-4207-8d14-28b375a25871/edit?view_items=bV8G0U69AJNc&invitationId=inv_d057a3b1-21d6-4290-9aea-5eb1c556a8ef#
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    TREE_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    key: builtins.str
    description: builtins.str
    @property
    def tree(self) -> global___Tree: ...
    type: global___FeatureType.ValueType
    def __init__(
        self,
        *,
        key: builtins.str = ...,
        description: builtins.str = ...,
        tree: global___Tree | None = ...,
        type: global___FeatureType.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["tree", b"tree"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["description", b"description", "key", b"key", "tree", b"tree", "type", b"type"]) -> None: ...

global___Feature = Feature

@typing_extensions.final
class Tree(google.protobuf.message.Message):
    """When the rules evaluator is traversing the tree, it will keep a local variable
    'value' that is updated along the way and is finally returned. It is initially
    set to the default value of the root node.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DEFAULT_FIELD_NUMBER: builtins.int
    CONSTRAINTS_FIELD_NUMBER: builtins.int
    DEFAULT_NEW_FIELD_NUMBER: builtins.int
    @property
    def default(self) -> google.protobuf.any_pb2.Any:
        """The default value to fall back to. If there are no constraints/rules
        defined, this is what gets returned.
        """
    @property
    def constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Constraint]: ...
    @property
    def default_new(self) -> global___Any: ...
    def __init__(
        self,
        *,
        default: google.protobuf.any_pb2.Any | None = ...,
        constraints: collections.abc.Iterable[global___Constraint] | None = ...,
        default_new: global___Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["default", b"default", "default_new", b"default_new"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["constraints", b"constraints", "default", b"default", "default_new", b"default_new"]) -> None: ...

global___Tree = Tree

@typing_extensions.final
class Constraint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RULE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    CONSTRAINTS_FIELD_NUMBER: builtins.int
    RULE_AST_FIELD_NUMBER: builtins.int
    RULE_AST_NEW_FIELD_NUMBER: builtins.int
    VALUE_NEW_FIELD_NUMBER: builtins.int
    rule: builtins.str
    """RulesLang string. Purely for readability. All edits to ruleslang
    are made through rule_ast_new instead.
    """
    @property
    def value(self) -> google.protobuf.any_pb2.Any:
        """This can be empty. If non-empty, and the above rule evaluated to true,
        then the rules engine should set its return value to this value.
        """
    @property
    def constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Constraint]:
        """If this list is empty, or none of the rules pass,
        return the most recent concrete value we traversed.
        """
    @property
    def rule_ast(self) -> lekko_client.gen.lekko.rules.v1beta2.rules_pb2.Rule:
        """Rules AST used for rules evaluation. It is a strict derivative of the
        string rule above.
        Deprecated: use rule_ast_new instead.
        """
    @property
    def rule_ast_new(self) -> lekko_client.gen.lekko.rules.v1beta3.rules_pb2.Rule:
        """Rules AST used for rules evaluation. It is an n-ary tree."""
    @property
    def value_new(self) -> global___Any: ...
    def __init__(
        self,
        *,
        rule: builtins.str = ...,
        value: google.protobuf.any_pb2.Any | None = ...,
        constraints: collections.abc.Iterable[global___Constraint] | None = ...,
        rule_ast: lekko_client.gen.lekko.rules.v1beta2.rules_pb2.Rule | None = ...,
        rule_ast_new: lekko_client.gen.lekko.rules.v1beta3.rules_pb2.Rule | None = ...,
        value_new: global___Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["rule_ast", b"rule_ast", "rule_ast_new", b"rule_ast_new", "value", b"value", "value_new", b"value_new"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["constraints", b"constraints", "rule", b"rule", "rule_ast", b"rule_ast", "rule_ast_new", b"rule_ast_new", "value", b"value", "value_new", b"value_new"]) -> None: ...

global___Constraint = Constraint

@typing_extensions.final
class Any(google.protobuf.message.Message):
    """New custom any type which allows us to manage dynamic types and values
    ourselves in application code.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TYPE_URL_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    type_url: builtins.str
    value: builtins.bytes
    def __init__(
        self,
        *,
        type_url: builtins.str = ...,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["type_url", b"type_url", "value", b"value"]) -> None: ...

global___Any = Any



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lekko/backend/v1beta1/distribution_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from lekko_client.gen.lekko.feature.v1beta1 import feature_pb2 as lekko_dot_feature_dot_v1beta1_dot_feature__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0lekko/backend/v1beta1/distribution_service.proto\x12\x15lekko.backend.v1beta1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a#lekko/feature/v1beta1/feature.proto\"K\n\rRepositoryKey\x12\x1d\n\nowner_name\x18\x01 \x01(\tR\townerName\x12\x1b\n\trepo_name\x18\x02 \x01(\tR\x08repoName\"\x7f\n\x1bGetRepositoryVersionRequest\x12?\n\x08repo_key\x18\x01 \x01(\x0b\x32$.lekko.backend.v1beta1.RepositoryKeyR\x07repoKey\x12\x1f\n\x0bsession_key\x18\x02 \x01(\tR\nsessionKey\"=\n\x1cGetRepositoryVersionResponse\x12\x1d\n\ncommit_sha\x18\x01 \x01(\tR\tcommitSha\"\xca\x01\n\x1cGetRepositoryContentsRequest\x12?\n\x08repo_key\x18\x01 \x01(\x0b\x32$.lekko.backend.v1beta1.RepositoryKeyR\x07repoKey\x12%\n\x0enamespace_name\x18\x02 \x01(\tR\rnamespaceName\x12!\n\x0c\x66\x65\x61ture_name\x18\x03 \x01(\tR\x0b\x66\x65\x61tureName\x12\x1f\n\x0bsession_key\x18\x04 \x01(\tR\nsessionKey\"\x80\x01\n\x1dGetRepositoryContentsResponse\x12\x1d\n\ncommit_sha\x18\x01 \x01(\tR\tcommitSha\x12@\n\nnamespaces\x18\x02 \x03(\x0b\x32 .lekko.backend.v1beta1.NamespaceR\nnamespaces\"[\n\tNamespace\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12:\n\x08\x66\x65\x61tures\x18\x02 \x03(\x0b\x32\x1e.lekko.backend.v1beta1.FeatureR\x08\x66\x65\x61tures\"i\n\x07\x46\x65\x61ture\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x10\n\x03sha\x18\x02 \x01(\tR\x03sha\x12\x38\n\x07\x66\x65\x61ture\x18\x03 \x01(\x0b\x32\x1e.lekko.feature.v1beta1.FeatureR\x07\x66\x65\x61ture\"2\n\nContextKey\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x12\n\x04type\x18\x02 \x01(\tR\x04type\"\x8f\x03\n\x13\x46lagEvaluationEvent\x12?\n\x08repo_key\x18\x01 \x01(\x0b\x32$.lekko.backend.v1beta1.RepositoryKeyR\x07repoKey\x12\x1d\n\ncommit_sha\x18\x02 \x01(\tR\tcommitSha\x12\x1f\n\x0b\x66\x65\x61ture_sha\x18\x03 \x01(\tR\nfeatureSha\x12%\n\x0enamespace_name\x18\x04 \x01(\tR\rnamespaceName\x12!\n\x0c\x66\x65\x61ture_name\x18\x05 \x01(\tR\x0b\x66\x65\x61tureName\x12\x44\n\x0c\x63ontext_keys\x18\x06 \x03(\x0b\x32!.lekko.backend.v1beta1.ContextKeyR\x0b\x63ontextKeys\x12\x1f\n\x0bresult_path\x18\x07 \x03(\x05R\nresultPath\x12\x46\n\x11\x63lient_event_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0f\x63lientEventTime\"\x87\x01\n SendFlagEvaluationMetricsRequest\x12\x42\n\x06\x65vents\x18\x01 \x03(\x0b\x32*.lekko.backend.v1beta1.FlagEvaluationEventR\x06\x65vents\x12\x1f\n\x0bsession_key\x18\x02 \x01(\tR\nsessionKey\"#\n!SendFlagEvaluationMetricsResponse\"\xdc\x01\n\x15RegisterClientRequest\x12?\n\x08repo_key\x18\x01 \x01(\x0b\x32$.lekko.backend.v1beta1.RepositoryKeyR\x07repoKey\x12%\n\x0enamespace_list\x18\x02 \x03(\tR\rnamespaceList\x12\x32\n\x15initial_bootstrap_sha\x18\x03 \x01(\tR\x13initialBootstrapSha\x12\'\n\x0fsidecar_version\x18\x04 \x01(\tR\x0esidecarVersion\"9\n\x16RegisterClientResponse\x12\x1f\n\x0bsession_key\x18\x04 \x01(\tR\nsessionKey\":\n\x17\x44\x65registerClientRequest\x12\x1f\n\x0bsession_key\x18\x01 \x01(\tR\nsessionKey\"\x1a\n\x18\x44\x65registerClientResponse\" \n\x1eGetDeveloperAccessTokenRequest\"7\n\x1fGetDeveloperAccessTokenResponse\x12\x14\n\x05token\x18\x01 \x01(\tR\x05token2\xa8\x06\n\x13\x44istributionService\x12\x81\x01\n\x14GetRepositoryVersion\x12\x32.lekko.backend.v1beta1.GetRepositoryVersionRequest\x1a\x33.lekko.backend.v1beta1.GetRepositoryVersionResponse\"\x00\x12\x84\x01\n\x15GetRepositoryContents\x12\x33.lekko.backend.v1beta1.GetRepositoryContentsRequest\x1a\x34.lekko.backend.v1beta1.GetRepositoryContentsResponse\"\x00\x12\x90\x01\n\x19SendFlagEvaluationMetrics\x12\x37.lekko.backend.v1beta1.SendFlagEvaluationMetricsRequest\x1a\x38.lekko.backend.v1beta1.SendFlagEvaluationMetricsResponse\"\x00\x12o\n\x0eRegisterClient\x12,.lekko.backend.v1beta1.RegisterClientRequest\x1a-.lekko.backend.v1beta1.RegisterClientResponse\"\x00\x12u\n\x10\x44\x65registerClient\x12..lekko.backend.v1beta1.DeregisterClientRequest\x1a/.lekko.backend.v1beta1.DeregisterClientResponse\"\x00\x12\x8a\x01\n\x17GetDeveloperAccessToken\x12\x35.lekko.backend.v1beta1.GetDeveloperAccessTokenRequest\x1a\x36.lekko.backend.v1beta1.GetDeveloperAccessTokenResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lekko.backend.v1beta1.distribution_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_REPOSITORYKEY']._serialized_start=145
  _globals['_REPOSITORYKEY']._serialized_end=220
  _globals['_GETREPOSITORYVERSIONREQUEST']._serialized_start=222
  _globals['_GETREPOSITORYVERSIONREQUEST']._serialized_end=349
  _globals['_GETREPOSITORYVERSIONRESPONSE']._serialized_start=351
  _globals['_GETREPOSITORYVERSIONRESPONSE']._serialized_end=412
  _globals['_GETREPOSITORYCONTENTSREQUEST']._serialized_start=415
  _globals['_GETREPOSITORYCONTENTSREQUEST']._serialized_end=617
  _globals['_GETREPOSITORYCONTENTSRESPONSE']._serialized_start=620
  _globals['_GETREPOSITORYCONTENTSRESPONSE']._serialized_end=748
  _globals['_NAMESPACE']._serialized_start=750
  _globals['_NAMESPACE']._serialized_end=841
  _globals['_FEATURE']._serialized_start=843
  _globals['_FEATURE']._serialized_end=948
  _globals['_CONTEXTKEY']._serialized_start=950
  _globals['_CONTEXTKEY']._serialized_end=1000
  _globals['_FLAGEVALUATIONEVENT']._serialized_start=1003
  _globals['_FLAGEVALUATIONEVENT']._serialized_end=1402
  _globals['_SENDFLAGEVALUATIONMETRICSREQUEST']._serialized_start=1405
  _globals['_SENDFLAGEVALUATIONMETRICSREQUEST']._serialized_end=1540
  _globals['_SENDFLAGEVALUATIONMETRICSRESPONSE']._serialized_start=1542
  _globals['_SENDFLAGEVALUATIONMETRICSRESPONSE']._serialized_end=1577
  _globals['_REGISTERCLIENTREQUEST']._serialized_start=1580
  _globals['_REGISTERCLIENTREQUEST']._serialized_end=1800
  _globals['_REGISTERCLIENTRESPONSE']._serialized_start=1802
  _globals['_REGISTERCLIENTRESPONSE']._serialized_end=1859
  _globals['_DEREGISTERCLIENTREQUEST']._serialized_start=1861
  _globals['_DEREGISTERCLIENTREQUEST']._serialized_end=1919
  _globals['_DEREGISTERCLIENTRESPONSE']._serialized_start=1921
  _globals['_DEREGISTERCLIENTRESPONSE']._serialized_end=1947
  _globals['_GETDEVELOPERACCESSTOKENREQUEST']._serialized_start=1949
  _globals['_GETDEVELOPERACCESSTOKENREQUEST']._serialized_end=1981
  _globals['_GETDEVELOPERACCESSTOKENRESPONSE']._serialized_start=1983
  _globals['_GETDEVELOPERACCESSTOKENRESPONSE']._serialized_end=2038
  _globals['_DISTRIBUTIONSERVICE']._serialized_start=2041
  _globals['_DISTRIBUTIONSERVICE']._serialized_end=2849
# @@protoc_insertion_point(module_scope)



import time
from threading import Event, Thread
from typing import Any, Dict, Optional

import grpc
from google.protobuf.any_pb2 import Any as ProtoAny

from lekko_client.clients.distribution_client import CachedDistributionClient
from lekko_client.exceptions import ClientNotInitialized
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    GetRepositoryContentsRequest,
    GetRepositoryContentsResponse,
    GetRepositoryVersionRequest,
)
from lekko_client.stores.store import Store


class CachedBackendClient(CachedDistributionClient):
    class RefreshThread(Thread):
        def __init__(self, client: "CachedBackendClient", refresh_interval_ms: int):
            super().__init__()
            self.daemon = True
            self.client = client
            self.refresh_interval = refresh_interval_ms
            self._enabled = True

        def stop(self):
            self._enabled = False

        def run(self):
            while self._enabled:
                if self.client.should_update_store():
                    self.client.update_store()
                time.sleep(self.refresh_interval / 1000)

    def __init__(
        self,
        uri: str,
        owner_name: str,
        repo_name: str,
        store: Store,
        api_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        credentials: grpc.ChannelCredentials = grpc.ssl_channel_credentials(),
        update_interval_ms: int = 1000,
    ):
        self.update_interval_ms = update_interval_ms
        self.timeout = None
        self.closed = False
        self.initialized_event = Event()
        super().__init__(uri, owner_name, repo_name, store, api_key, context, credentials)

    def initialize(self):
        super().initialize()
        self.refresh_thread = CachedBackendClient.RefreshThread(self, self.update_interval_ms)
        self.refresh_thread.start()

    def load_contents(self) -> Optional[GetRepositoryContentsResponse]:
        if not self._client:
            return None
        return self._client.GetRepositoryContents(
            GetRepositoryContentsRequest(repo_key=self.repository, session_key=self.session_key)
        )

    def update_store(self):
        self.load()
        self.initialized_event.set()

    def should_update_store(self) -> bool:
        if not self._client:
            return False
        if not self.initialized_event.is_set():
            return True
        version_response = self._client.GetRepositoryVersion(
            GetRepositoryVersionRequest(repo_key=self.repository, session_key=self.session_key)
        )
        current_sha = self.store.commit_sha
        return current_sha != version_response.commit_sha

    def close(self):
        super().close()
        self.refresh_thread.stop()
        self.initialized_event.clear()

    def get(self, namespace: str, key: str, context: Dict[str, Any]) -> ProtoAny:
        if not self.initialized_event.wait(timeout=5):  # Give the background thread 5 seconds to populate
            raise ClientNotInitialized("Repository contents not yet loaded from server")
        return super().get(namespace, key, context)



import json
import os
from unittest import mock

import grpc
import pytest
from google.protobuf import wrappers_pb2
from google.protobuf.any_pb2 import Any

from lekko_client.clients import APIClient, SidecarClient
from lekko_client.exceptions import (
    AuthenticationError,
    FeatureNotFound,
    MismatchedProtoType,
    MismatchedType,
)
from lekko_client.gen.lekko.client.v1beta1 import configuration_service_pb2 as messages


@pytest.mark.parametrize(
    "proto_fn_name,response_obj,response_val,test_fn_name",
    [
        ("GetBoolValue", messages.GetBoolValueResponse, True, "get_bool"),
        ("GetIntValue", messages.GetIntValueResponse, 10, "get_int"),
        ("GetFloatValue", messages.GetFloatValueResponse, 2.5, "get_float"),
        ("GetStringValue", messages.GetStringValueResponse, "test string", "get_string"),
    ],
)
def test_get_scalar(test_server, proto_fn_name, response_obj, response_val, test_fn_name):
    requests = [
        test_server.MockRequestResponse("Register", messages.RegisterResponse),
        test_server.MockRequestResponse(proto_fn_name, response_obj(value=response_val)),
    ]

    async_requests = test_server.mock_async_responses(requests)

    owner_name = "owner"
    repo_name = "repo"
    namespace = "namespace"
    feature_name = "val"
    api_key = "lekko_apikey123"
    context = {"ctx_key": response_val}

    client = SidecarClient(owner_name, repo_name, api_key)
    resp = getattr(client, test_fn_name)(namespace, feature_name, context)

    assert resp == response_val
    completed_requests = async_requests.result()
    assert len(completed_requests) == 2
    assert completed_requests[0].arg.repo_key.owner_name == owner_name
    assert completed_requests[0].arg.repo_key.repo_name == repo_name
    assert completed_requests[0].arg.namespace_list == []
    assert ("apikey", api_key) in completed_requests[0].metadata
    assert completed_requests[1].arg.key == feature_name
    assert completed_requests[1].arg.namespace == namespace
    req_ctx = {k: getattr(v, v.WhichOneof("kind")) for k, v in completed_requests[1].arg.context.items()}
    assert req_ctx == context
    assert ("apikey", api_key) in completed_requests[1].metadata


def test_get_json(test_server):
    expected = {"key": "value", "int_key": 1}
    requests = [
        test_server.MockRequestResponse("Register", messages.RegisterResponse),
        test_server.MockRequestResponse(
            "GetJSONValue", messages.GetJSONValueResponse(value=bytes(json.dumps(expected), "utf-8"))
        ),
    ]
    test_server.mock_async_responses(requests)

    client = SidecarClient("owner", "repo", "lekko_apikey123")
    resp = client.get_json("val", "namespace", {})

    assert resp == expected


def test_get_proto_by_type(test_server):
    any_proto = Any()
    int_proto = wrappers_pb2.Int32Value(value=10)
    any_proto.Pack(int_proto)
    requests = [
        test_server.MockRequestResponse("Register", messages.RegisterResponse),
        test_server.MockRequestResponse("GetProtoValue", messages.GetProtoValueResponse(value=any_proto)),
        test_server.MockRequestResponse("GetProtoValue", messages.GetProtoValueResponse(value=any_proto)),
    ]
    test_server.mock_async_responses(requests)

    client = SidecarClient("owner", "repo", "lekko_apikey123")
    resp = client.get_proto_by_type("val", "namespace", {}, wrappers_pb2.Int32Value)

    assert resp == int_proto

    with pytest.raises(MismatchedProtoType):
        client.get_proto_by_type("val", "namespace", {}, wrappers_pb2.Int64Value)


def test_get_proto(test_server):
    any_proto = Any()
    int_proto = wrappers_pb2.Int32Value(value=10)
    any_proto.Pack(int_proto)
    lekko_any_proto = messages.Any(type_url=any_proto.type_url, value=any_proto.value)
    requests = [
        test_server.MockRequestResponse("Register", messages.RegisterResponse),
        test_server.MockRequestResponse("GetProtoValue", messages.GetProtoValueResponse(value=any_proto)),
        test_server.MockRequestResponse("GetProtoValue", messages.GetProtoValueResponse(value=any_proto)),
        test_server.MockRequestResponse(
            "GetProtoValue", messages.GetProtoValueResponse(value=any_proto, value_v2=lekko_any_proto)
        ),
        test_server.MockRequestResponse("GetProtoValue", messages.GetProtoValueResponse(value_v2=lekko_any_proto)),
    ]
    test_server.mock_async_responses(requests)

    client = SidecarClient("owner", "repo", "lekko_apikey123")

    # When the proto symbol is loaded in the db, it should Unpack correctly
    resp = client.get_proto("val", "namespace", {})
    assert resp == int_proto

    # If the proto symbol can't be found, we fall through to returning the Any proto
    with mock.patch("google.protobuf.symbol_database.SymbolDatabase.GetSymbol", side_effect=KeyError):
        resp = client.get_proto("val", "namespace", {})
        assert resp == any_proto

    # test get proto with value and value_v2 being returned
    resp = client.get_proto("val", "namespace", {})
    assert resp == int_proto

    # test get proto with only value_v2 being returned
    resp = client.get_proto("val", "namespace", {})
    assert resp == int_proto


def test_missing_api_key(test_server):
    expected = True
    requests = [
        test_server.MockRequestResponse("Register", messages.RegisterResponse),
        test_server.MockRequestResponse("GetBoolValue", messages.GetBoolValueResponse(value=expected)),
    ]
    test_server.mock_async_responses(requests)

    with pytest.raises(AuthenticationError):
        SidecarClient("owner", "repo")

    os.environ["LEKKO_API_KEY"] = "lekko_apikey123"
    client = SidecarClient("owner", "repo")
    result = client.get_bool("key", "namespace", {})
    assert result == expected


# Interceptor can't handle returning errors for some reason
def test_errors(test_server_no_interceptor):
    requests = [
        test_server_no_interceptor.MockRequestResponse("Register", messages.RegisterResponse),
        test_server_no_interceptor.MockRequestResponse(
            fn_name="GetBoolValue",
            response=None,
            status_code=grpc.StatusCode.NOT_FOUND,
            error_text="get evaluable feature: first feature: record not found",
        ),
        test_server_no_interceptor.MockRequestResponse(
            "GetBoolValue",
            None,
            grpc.StatusCode.INVALID_ARGUMENT,
            "requested feature is not of type bool",
        ),
    ]

    test_server_no_interceptor.mock_async_responses(requests)

    client = SidecarClient("owner", "repo", "lekko_apikey123")
    with pytest.raises(FeatureNotFound):
        client.get_bool("key", "namespace", {})

    with pytest.raises(MismatchedType):
        client.get_bool("key", "namespace", {})


def test_proto_errors(test_server_no_interceptor):
    requests = [
        test_server_no_interceptor.MockRequestResponse("Register", messages.RegisterResponse),
        test_server_no_interceptor.MockRequestResponse(
            fn_name="GetProtoValue",
            response=None,
            status_code=grpc.StatusCode.NOT_FOUND,
            error_text="get evaluable feature: first feature: record not found",
        ),
        test_server_no_interceptor.MockRequestResponse(
            "GetProtoValue",
            None,
            grpc.StatusCode.INVALID_ARGUMENT,
            "requested feature is not of type proto",
        ),
        test_server_no_interceptor.MockRequestResponse(
            "GetProtoValue",
            None,
            grpc.StatusCode.INTERNAL,
            "internal server error",
        ),
    ]

    test_server_no_interceptor.mock_async_responses(requests)

    client = SidecarClient("owner", "repo", "lekko_apikey123")
    with pytest.raises(FeatureNotFound):
        client.get_proto("key", "namespace", {})

    with pytest.raises(MismatchedType):
        client.get_proto("key", "namespace", {})

    with pytest.raises(grpc.RpcError):
        client.get_proto("key", "namespace", {})


def test_get_api_client(test_server):
    requests = [
        test_server.MockRequestResponse("Register", messages.RegisterResponse),
        test_server.MockRequestResponse("GetStringValue", messages.GetStringValueResponse(value="feature value")),
    ]
    test_server.mock_async_responses(requests)

    client = APIClient("owner", "repo", "lekko_apikey123")
    resp = client.get_string("val", "namespace", {"conext": "hello"})

    assert resp == "feature value"



from typing import List

import pytest
from google.protobuf.wrappers_pb2 import Int64Value

from lekko_client.evaluation.evaluation import evaluate
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import Feature
from lekko_client.helpers import convert_context
from lekko_client.models import ClientContext


@pytest.mark.parametrize(
    "context,expected",
    [
        ({"a": 1}, 0),
        ({"a": 1, "c": 3}, 1),
        ({"c": 3}, 2),
        ({"f": 3}, 0),
        ({"f": 5}, 5),
        ({"h": 4}, 7),
        ({"i": 4}, 8),
        ({"p": "a foo bar"}, 15),
        ({"q": "hello world"}, 16),
        ({"r": "a foo bar"}, 17),
        ({"s": 2}, 18),
        ({"t": "anything"}, 19),
        ({"u": 11}, 20),
    ],
)
def test_complex_evaluation(test_complex_rule_feature, context, expected):
    client_context = convert_context(context)
    result = evaluate(test_complex_rule_feature, "default", client_context)
    inner_result = Int64Value()
    assert result.value.Unpack(inner_result)
    assert inner_result.value == expected


def test_empty_config_tree():
    with pytest.raises(Exception):
        evaluate(Feature(), "ns")


@pytest.mark.parametrize(
    "feature_fixture_name,test_context,expected_fixture_name,expected_path",
    [
        ("test_feature_no_constraints", convert_context({}), "test_feature_default_value", []),
        ("test_feature_no_constraints", convert_context({"key": "anything"}), "test_feature_default_value", []),
        ("test_feature_no_constraints", None, "test_feature_default_value", []),
        ("test_feature_one_level_traversal", convert_context({}), "test_feature_default_value", []),
        ("test_feature_one_level_traversal", convert_context({"age": 5}), "test_feature_default_value", []),
        ("test_feature_one_level_traversal", convert_context({"age": 10}), "test_feature_constraint_value", [0]),
        ("test_feature_one_level_traversal", convert_context({"age": 12}), "test_feature_constraint_value", [1]),
        ("test_feature_two_level_traversal", convert_context({}), "test_feature_default_value", []),
        ("test_feature_two_level_traversal", convert_context({"age": 10}), "test_feature_constraint_value", [0]),
        (
            "test_feature_two_level_traversal",
            convert_context({"age": 10, "city": "Rome"}),
            "test_feature_constraint_value",
            [0, 0],
        ),
        (
            "test_feature_two_level_traversal",
            convert_context({"age": 10, "city": "Paris"}),
            "test_feature_constraint_value",
            [0, 1],
        ),
        (
            "test_feature_two_level_traversal",
            convert_context({"age": 12, "city": "Milan"}),
            "test_feature_constraint_value",
            [1],
        ),
        ("test_feature_two_level_traversal", convert_context({"age": 12}), "test_feature_constraint_value", [1]),
        (
            "test_feature_two_level_traversal",
            convert_context({"age": 12, "city": "Rome"}),
            "test_feature_constraint_value",
            [1, 0],
        ),
        (
            "test_feature_two_level_traversal",
            convert_context({"age": 12, "city": "Paris"}),
            "test_feature_constraint_value",
            [1, 1],
        ),
        (
            "test_feature_two_level_traversal",
            convert_context({"age": 12, "city": "Milan"}),
            "test_feature_constraint_value",
            [1],
        ),
    ],
)
def test_eval(
    feature_fixture_name: str,
    test_context: ClientContext,
    expected_fixture_name: str,
    expected_path: List[int],
    request,
):
    test_feature = request.getfixturevalue(feature_fixture_name)
    expected = request.getfixturevalue(expected_fixture_name)
    if expected is not None:
        result = evaluate(test_feature, "ns_1", test_context)
        assert result.value == expected
        assert result.path == expected_path
    else:
        assert not "test case needs to either expect an error or a result"



import concurrent.futures.thread
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple
from unittest import mock

import grpc
import grpc_testing
import pytest
from google.protobuf.any_pb2 import Any as ProtoAny
from google.protobuf.message import Message as ProtoMessage
from google.protobuf.struct_pb2 import Struct, Value
from google.protobuf.wrappers_pb2 import Int64Value
from grpc_testing import _channel  # noqa

from lekko_client import helpers
from lekko_client.clients.distribution_client import CachedDistributionClient
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    GetRepositoryContentsResponse,
)
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2_grpc import (
    DistributionServiceStub,
)
from lekko_client.gen.lekko.client.v1beta1.configuration_service_pb2 import DESCRIPTOR
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import Any as LekkoAny
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import (
    Constraint,
    Feature,
    FeatureType,
    Tree,
)
from lekko_client.gen.lekko.rules.v1beta3.rules_pb2 import (
    Atom,
    ComparisonOperator,
    Rule,
)
from lekko_client.stores.store import Store


@pytest.fixture
def test_thread():
    client_execution_thread_pool = ThreadPoolExecutor(max_workers=2)

    try:
        yield client_execution_thread_pool
    finally:
        client_execution_thread_pool.shutdown(wait=False)
        client_execution_thread_pool._threads.clear()
        concurrent.futures.thread._threads_queues.clear()


@pytest.fixture
def test_channel_no_interceptor(test_thread):
    channel = grpc_testing.channel(DESCRIPTOR.services_by_name.values(), grpc_testing.strict_real_time())
    try:
        with mock.patch("lekko_client.clients.config_client.get_grpc_channel", return_value=channel):
            yield channel
    finally:
        channel.close()


@pytest.fixture
def test_channel(test_thread):
    channel = grpc_testing.channel(DESCRIPTOR.services_by_name.values(), grpc_testing.strict_real_time())
    try:
        with mock.patch("grpc.insecure_channel", return_value=channel), mock.patch(
            "grpc.secure_channel", return_value=channel
        ):
            yield channel
    finally:
        channel.close()


class PatchedUnaryUnary(grpc_testing._channel._multi_callable.UnaryUnary):
    # grpc_testing is very out of date and its fn sigs don't take wait_for_ready or compression
    def with_call(self, request, timeout=None, metadata=None, credentials=None, wait_for_ready=None, compression=None):
        return super().with_call(request, timeout, metadata, credentials)


grpc_testing._channel._channel._multi_callable.UnaryUnary = PatchedUnaryUnary


class MockServer:
    def __init__(self, channel, thread):
        self.channel = channel
        self.thread = thread
        helpers._CHANNELS = {}

    @dataclass
    class MockRequestResponse:
        fn_name: str
        response: ProtoMessage
        status_code: grpc.StatusCode = grpc.StatusCode.OK
        error_text: str = ""

    @dataclass
    class CompletedRequest:
        arg: ProtoMessage
        metadata: Tuple[Tuple[str, str]]

    def mock_async_responses(self, mock_requests: List[MockRequestResponse]):
        def server() -> List[MockServer.CompletedRequest]:
            service = DESCRIPTOR.services_by_name["ConfigurationService"]
            incoming_requests = []
            for mock_request in mock_requests:
                metadata, request, rpc = self.channel.take_unary_unary(service.methods_by_name[mock_request.fn_name])
                rpc.terminate(mock_request.response, [], mock_request.status_code, mock_request.error_text)
                incoming_requests.append(MockServer.CompletedRequest(arg=request, metadata=metadata))

            return incoming_requests

        server_future = self.thread.submit(server)
        return server_future


@pytest.fixture
def test_server(test_channel, test_thread):
    return MockServer(test_channel, test_thread)


@pytest.fixture
def test_server_no_interceptor(test_channel_no_interceptor, test_thread):
    return MockServer(test_channel_no_interceptor, test_thread)


@pytest.fixture
def test_complex_rule_feature():
    filename = "tests/fixtures/rules.proto.bin"
    feature = Feature()
    with open(filename, "rb") as f:
        feature.ParseFromString(f.read())

    return feature


@pytest.fixture
def test_feature_default_value() -> ProtoAny:
    any_proto = ProtoAny()
    any_proto.Pack(Int64Value(value=1))
    return any_proto


@pytest.fixture
def test_feature_constraint_value() -> ProtoAny:
    any_proto = ProtoAny()
    any_proto.Pack(Int64Value(value=2))
    return any_proto


def convert_to_value(v: Any) -> Value:
    s = Struct()
    s.update({"key": v})
    return s.fields["key"]


@pytest.fixture
def test_feature_no_constraints(test_feature_default_value) -> Feature:
    return Feature(
        key="key",
        description="config description",
        type=FeatureType.FEATURE_TYPE_INT,
        tree=Tree(
            default=test_feature_default_value,
            default_new=LekkoAny(type_url=test_feature_default_value.type_url, value=test_feature_default_value.value),
        ),
    )


@pytest.fixture
def test_feature_one_level_traversal(test_feature_default_value, test_feature_constraint_value) -> Feature:
    return Feature(
        key="key",
        description="config description",
        type=FeatureType.FEATURE_TYPE_INT,
        tree=Tree(
            default=test_feature_default_value,
            default_new=LekkoAny(type_url=test_feature_default_value.type_url, value=test_feature_default_value.value),
            constraints=[
                Constraint(
                    rule_ast_new=Rule(
                        atom=Atom(
                            context_key="age",
                            comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                            comparison_value=convert_to_value(10),
                        )
                    ),
                    value=test_feature_constraint_value,
                    value_new=LekkoAny(
                        type_url=test_feature_constraint_value.type_url, value=test_feature_constraint_value.value
                    ),
                ),
                Constraint(
                    rule_ast_new=Rule(
                        atom=Atom(
                            context_key="age",
                            comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                            comparison_value=convert_to_value(12),
                        )
                    ),
                    value=test_feature_constraint_value,
                    value_new=LekkoAny(
                        type_url=test_feature_constraint_value.type_url, value=test_feature_constraint_value.value
                    ),
                ),
            ],
        ),
    )


@pytest.fixture
def test_feature_two_level_traversal(test_feature_default_value, test_feature_constraint_value) -> Feature:
    return Feature(
        key="key",
        description="config description",
        type=FeatureType.FEATURE_TYPE_INT,
        tree=Tree(
            default=test_feature_default_value,
            default_new=LekkoAny(type_url=test_feature_default_value.type_url, value=test_feature_default_value.value),
            constraints=[
                Constraint(
                    rule_ast_new=Rule(
                        atom=Atom(
                            context_key="age",
                            comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                            comparison_value=convert_to_value(10),
                        )
                    ),
                    value=test_feature_constraint_value,
                    value_new=LekkoAny(
                        type_url=test_feature_constraint_value.type_url, value=test_feature_constraint_value.value
                    ),
                    constraints=[
                        Constraint(
                            rule_ast_new=Rule(
                                atom=Atom(
                                    context_key="city",
                                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                                    comparison_value=convert_to_value("Rome"),
                                )
                            ),
                            value=test_feature_constraint_value,
                            value_new=LekkoAny(
                                type_url=test_feature_constraint_value.type_url,
                                value=test_feature_constraint_value.value,
                            ),
                        ),
                        Constraint(
                            rule_ast_new=Rule(
                                atom=Atom(
                                    context_key="city",
                                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                                    comparison_value=convert_to_value("Paris"),
                                )
                            ),
                            value=test_feature_constraint_value,
                            value_new=LekkoAny(
                                type_url=test_feature_constraint_value.type_url,
                                value=test_feature_constraint_value.value,
                            ),
                        ),
                    ],
                ),
                Constraint(
                    rule_ast_new=Rule(
                        atom=Atom(
                            context_key="age",
                            comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                            comparison_value=convert_to_value(12),
                        )
                    ),
                    value=test_feature_constraint_value,
                    value_new=LekkoAny(
                        type_url=test_feature_constraint_value.type_url, value=test_feature_constraint_value.value
                    ),
                    constraints=[
                        Constraint(
                            rule_ast_new=Rule(
                                atom=Atom(
                                    context_key="city",
                                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                                    comparison_value=convert_to_value("Rome"),
                                )
                            ),
                            value=test_feature_constraint_value,
                            value_new=LekkoAny(
                                type_url=test_feature_constraint_value.type_url,
                                value=test_feature_constraint_value.value,
                            ),
                        ),
                        Constraint(
                            rule_ast_new=Rule(
                                atom=Atom(
                                    context_key="city",
                                    comparison_operator=ComparisonOperator.COMPARISON_OPERATOR_EQUALS,
                                    comparison_value=convert_to_value("Paris"),
                                )
                            ),
                            value=test_feature_constraint_value,
                            value_new=LekkoAny(
                                type_url=test_feature_constraint_value.type_url,
                                value=test_feature_constraint_value.value,
                            ),
                        ),
                    ],
                ),
            ],
        ),
    )


@pytest.fixture
def mock_store() -> Store:
    store = mock.Mock(spec_set=Store)
    type(store).commit_sha = mock.PropertyMock(return_value="test_commit_sha")
    return store


@pytest.fixture
def mock_event_batcher():
    def mock_event_batcher_factory(self, client, session_key, interval):
        return mock.Mock(spec_set=CachedDistributionClient.EventsBatcher)

    return mock_event_batcher_factory


@pytest.fixture
def mock_distribution_client(mock_store, mock_event_batcher) -> CachedDistributionClient:
    class MockDistributionClient(CachedDistributionClient):
        EventsBatcher = mock_event_batcher

        def initialize(self):
            pass

        def load_contents(self) -> Optional[GetRepositoryContentsResponse]:
            return mock.Mock(spec_set=GetRepositoryContentsResponse)

        def initialize_client(self, credentials: grpc.ChannelCredentials):
            self._client = mock.Mock(spec_set=DistributionServiceStub)
            self.session_key = "session key"

    client = MockDistributionClient("uri", "owner", "repo", mock_store, api_key="api_key")
    return client



import json
import logging
import queue
import time
from abc import abstractmethod
from datetime import datetime
from threading import Thread
from typing import Any, Dict, List, Optional, Type, TypeVar

import grpc
from google.protobuf import descriptor_pool as proto_descriptor_pool
from google.protobuf import symbol_database as proto_symbol_database
from google.protobuf.any_pb2 import Any as ProtoAny
from google.protobuf.json_format import MessageToJson
from google.protobuf.message import Message as ProtoMessage
from google.protobuf.struct_pb2 import Value
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import BoolValue, FloatValue, Int64Value, StringValue

from lekko_client.clients.client import Client
from lekko_client.evaluation.evaluation import EvaluationResult, evaluate
from lekko_client.exceptions import LekkoRpcError, MismatchedProtoType, MismatchedType
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    DeregisterClientRequest,
    FlagEvaluationEvent,
    GetRepositoryContentsResponse,
    RegisterClientRequest,
    RepositoryKey,
    SendFlagEvaluationMetricsRequest,
)
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2_grpc import (
    DistributionServiceStub,
)
from lekko_client.helpers import convert_context, get_context_keys, get_grpc_channel
from lekko_client.models import ClientContext, FeatureData
from lekko_client.stores.store import Store

log = logging.getLogger(__name__)


class CachedDistributionClient(Client):
    class EventsBatcher(Thread):
        def __init__(
            self, dist_client: DistributionServiceStub, session_key: str, upload_interval_ms: int, batch_size: int
        ):
            super().__init__()
            self.daemon = True
            self.dist_client = dist_client
            self.upload_interval = upload_interval_ms / 1000
            self.batch_size = batch_size
            self.session_key = session_key
            self.events: List[FlagEvaluationEvent] = []
            self._enabled = True
            self.queue: queue.Queue[FlagEvaluationEvent] = queue.Queue()

        def stop(self):
            self._enabled = False

        def add_event(self, event: FlagEvaluationEvent):
            self.queue.put(event)

        def upload_events(self):
            if self.events:
                self.dist_client.SendFlagEvaluationMetrics(
                    SendFlagEvaluationMetricsRequest(events=self.events, session_key=self.session_key)
                )

        def run(self):
            last_upload_time = time.time()
            while self._enabled:
                event = self.queue.get()
                self.events.append(event)
                now = time.time()
                if (len(self.events) > self.batch_size) or (now - last_upload_time > self.upload_interval):
                    try:
                        self.upload_events()
                        last_upload_time = now
                    except Exception:
                        log.warning("Failed to upload config evaluation events.")

    def __init__(
        self,
        uri: str,
        owner_name: str,
        repo_name: str,
        store: Store,
        api_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        credentials: grpc.ChannelCredentials = grpc.ssl_channel_credentials(),
    ):
        super().__init__(owner_name, repo_name, api_key, context)
        self.uri = uri
        self.repository = RepositoryKey(owner_name=owner_name, repo_name=repo_name)
        self.store = store
        self._client: Optional[DistributionServiceStub] = None
        self.session_key = ""
        self.events_batcher = None
        if self.api_key:
            self.initialize_client(credentials)
            if self._client:
                self.events_batcher = self.EventsBatcher(
                    self._client, self.session_key, upload_interval_ms=5 * 1_000, batch_size=100
                )
                self.events_batcher.start()

        self.initialize()

    _TYPE_MAPPING: Dict[Type, Type[BoolValue | Int64Value | StringValue | FloatValue]] = {
        bool: BoolValue,
        int: Int64Value,
        str: StringValue,
        float: FloatValue,
    }

    def initialize_client(self, credentials: grpc.ChannelCredentials):
        from lekko_client import __version__

        channel = get_grpc_channel(self.uri, self.api_key, credentials)
        self._client = DistributionServiceStub(channel)
        try:
            register_response = self._client.RegisterClient(
                RegisterClientRequest(repo_key=self.repository, sidecar_version=__version__)
            )
        except grpc.RpcError as e:
            raise LekkoRpcError(f"Unable to register distribution service: {e}")
        self.session_key = register_response.session_key

    @abstractmethod
    def initialize(self) -> None:
        ...

    def load(self) -> bool:
        contents = self.load_contents()
        if not contents:
            return False
        loaded = self.store.load(contents)
        return loaded

    @abstractmethod
    def load_contents(self) -> Optional[GetRepositoryContentsResponse]:
        ...

    def track(
        self, namespace: str, feature_data: FeatureData, result: EvaluationResult, context: Optional[ClientContext]
    ) -> None:
        if not self.events_batcher:
            return

        timestamp = Timestamp()
        timestamp.FromDatetime(datetime.utcnow())
        event = FlagEvaluationEvent(
            repo_key=self.repository,
            commit_sha=self.store.commit_sha,
            feature_sha=feature_data.config_sha,
            namespace_name=namespace,
            feature_name=feature_data.feature.key,
            context_keys=get_context_keys(context),
            result_path=result.path,
            client_event_time=timestamp,
        )
        self.events_batcher.add_event(event)

    def get(self, namespace: str, key: str, context: Dict[str, Any]) -> ProtoAny:
        feature_data = self.store.get(namespace, key)
        client_context = convert_context(context)
        result = evaluate(feature_data.feature, namespace, client_context)
        self.track(namespace, feature_data, result, client_context)
        return result.value

    ReturnType = TypeVar("ReturnType", str, float, int, bool)

    def get_scalar(self, namespace: str, key: str, context: Dict[str, Any], typ: Type[ReturnType]) -> ReturnType:
        result = self.get(namespace, key, context)
        return_wrapper = self._TYPE_MAPPING[typ]()
        if result.Unpack(return_wrapper):
            return return_wrapper.value  # type:ignore
        raise MismatchedType(f"Feature {key} is of type {result.type_url} and cannot be converted to {typ}")

    def get_bool(self, namespace: str, key: str, context: Dict[str, Any]) -> bool:
        return self.get_scalar(namespace, key, context, bool)

    def get_int(self, namespace: str, key: str, context: Dict[str, Any]) -> int:
        return self.get_scalar(namespace, key, context, int)

    def get_float(self, namespace: str, key: str, context: Dict[str, Any]) -> float:
        return self.get_scalar(namespace, key, context, float)

    def get_string(self, namespace: str, key: str, context: Dict[str, Any]) -> str:
        return self.get_scalar(namespace, key, context, str)

    def get_json(self, namespace: str, key: str, context: Dict[str, Any]) -> Any:
        result = self.get(namespace, key, context)
        return_wrapper = Value()
        result.Unpack(return_wrapper)
        return json.loads(MessageToJson(return_wrapper))

    def get_proto(self, namespace: str, key: str, context: Dict[str, Any]) -> ProtoMessage:
        val = self.get(namespace, key, context)
        db = proto_symbol_database.SymbolDatabase(pool=proto_descriptor_pool.Default())
        try:
            ret_val = db.GetSymbol(val.type_url.split("/")[1])()
            if val.Unpack(ret_val):
                return ret_val
        except (KeyError, IndexError):
            pass
        return val

    def get_proto_by_type(
        self,
        namespace: str,
        key: str,
        context: Dict[str, Any],
        proto_message_type: Type[Client.ProtoType],
    ) -> Client.ProtoType:
        val = self.get(namespace, key, context)
        ret_val = proto_message_type()
        if val.Unpack(ret_val):
            return ret_val

        raise MismatchedProtoType(f"Error unpacking from {val.type_url} to {proto_message_type.DESCRIPTOR.name}")

    def close(self):
        super().close()
        if self._client and self.session_key:
            if self.events_batcher:
                self.events_batcher.upload_events()
                self.events_batcher.stop()
                self.events_batcher.join(timeout=5)
            self._client.DeregisterClient(DeregisterClientRequest(session_key=self.session_key))



import glob
import os
from typing import Any, Dict, List, Optional

import grpc
import yaml
from dulwich.errors import NotGitRepository
from dulwich.object_store import tree_lookup_path
from dulwich.repo import Repo as GitRepo
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver

from lekko_client.clients.distribution_client import CachedDistributionClient
from lekko_client.exceptions import GitRepoNotFound
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    Feature as DistFeature,
)
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    GetRepositoryContentsResponse,
    Namespace,
)
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import Feature
from lekko_client.stores.store import Store


class CachedGitClient(CachedDistributionClient):
    ROOT_CONFIG_METADATA_FILENAME = "lekko.root.yaml"

    class GitFileEventHandler(FileSystemEventHandler):
        def __init__(self, client: "CachedGitClient") -> None:
            super().__init__()
            self.client = client

        def on_any_event(self, event):
            self.client.load()
            return super().on_any_event(event)

    def __init__(
        self,
        lekko_uri: str,
        repository_owner: str,
        repository_name: str,
        store: Store,
        path: str,
        api_key: Optional[str],
        context: Optional[Dict[str, Any]] = None,
        credentials: grpc.ChannelCredentials = grpc.ssl_channel_credentials(),
        should_watch: Optional[bool] = True,
    ):
        super().__init__(lekko_uri, repository_owner, repository_name, store, api_key, context, credentials)
        self.watcher: Optional[BaseObserver] = None
        self.path = path
        self.should_watch = should_watch

    def initialize(self) -> None:
        super().initialize()
        self.load()
        if self.should_watch:
            event_handler = CachedGitClient.GitFileEventHandler(self)
            self.watcher = Observer()
            self.watcher.schedule(event_handler, self.path, recursive=True)
            self.watcher.start()

    def load_contents(self) -> GetRepositoryContentsResponse:
        try:
            repo = GitRepo(self.path)
        except NotGitRepository:
            raise GitRepoNotFound(f"{self.path} is not a git repository")

        return GetRepositoryContentsResponse(
            commit_sha=repo.head().decode("utf-8"), namespaces=self.get_namespaces(repo)
        )

    def get_namespaces(self, repo: GitRepo) -> List[Namespace]:
        md_file_path = os.path.join(self.path, self.ROOT_CONFIG_METADATA_FILENAME)
        with open(md_file_path) as f:
            md_contents = yaml.safe_load(f)

        ns_names = md_contents.get("namespaces", [])
        return [Namespace(name=ns_name, features=self.get_configs(repo, ns_name)) for ns_name in ns_names]

    def get_configs(self, repo: GitRepo, ns_name: str) -> List[DistFeature]:
        proto_dir_path = os.path.join(self.path, ns_name, "gen", "proto")
        if not os.path.isdir(proto_dir_path):
            return []

        features = []
        for proto_bin_file in glob.glob(os.path.join(proto_dir_path, "*.proto.bin")):
            proto_bin_relative_filename = os.path.relpath(proto_bin_file, self.path)
            with open(proto_bin_file, "rb") as proto_bin:
                _, sha = tree_lookup_path(
                    repo.get_object, repo[repo.head()].tree, proto_bin_relative_filename.encode()  # type: ignore
                )
                feature = Feature()
                feature.ParseFromString(proto_bin.read())
                features.append(
                    DistFeature(
                        name=os.path.basename(proto_bin_file).replace(".proto.bin", ""),
                        sha=sha.decode("utf-8"),
                        feature=feature,
                    )
                )
        return features

    def close(self):
        super().close()
        # if self.events_batcher:
        # await self.events_batcher.close()
        if self.watcher:
            self.watcher.stop()
            self.watcher.join()



import struct

from google.protobuf.struct_pb2 import Value
from xxhash import xxh32

from lekko_client.exceptions import EvaluationError
from lekko_client.gen.lekko.client.v1beta1.configuration_service_pb2 import (
    Value as LekkoValue,
)
from lekko_client.gen.lekko.rules.v1beta3.rules_pb2 import (
    CallExpression,
    ComparisonOperator,
    LogicalOperator,
    Rule,
)
from lekko_client.models import ClientContext


def evaluate_rule(rule: Rule, namespace: str, config_name: str, context: ClientContext = None) -> bool:
    if not rule:
        raise EvaluationError("Empty rule")

    rule_type = rule.WhichOneof("rule")
    if not rule_type:
        raise EvaluationError("Empty rule")

    rule_value = getattr(rule, rule_type)

    if rule_type == "bool_const":
        return rule_value
    elif rule_type == "not":
        return not evaluate_rule(rule_value, namespace, config_name, context)
    elif rule_type == "logical_expression":
        if not rule_value.rules:
            raise EvaluationError("No rules found in logical expression")

        logical_operator = rule_value.logical_operator
        return (
            all(evaluate_rule(r, namespace, config_name, context) for r in rule_value.rules)
            if logical_operator == LogicalOperator.LOGICAL_OPERATOR_AND
            else any(evaluate_rule(r, namespace, config_name, context) for r in rule_value.rules)
        )
    elif rule_type == "atom":
        context_key = rule_value.context_key
        context_value = context.get(context_key) if context else None

        if rule_value.comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_PRESENT:
            return context_value is not None

        if context_value is None:
            return False

        if rule_value.comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_EQUALS:
            return evaluate_equals(rule_value.comparison_value, context_value)
        elif rule_value.comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_NOT_EQUALS:
            return not evaluate_equals(rule_value.comparison_value, context_value)
        elif rule_value.comparison_operator in (
            ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN,
            ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN_OR_EQUALS,
            ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN,
            ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN_OR_EQUALS,
        ):
            return evaluate_number_comparator(
                rule_value.comparison_operator, rule_value.comparison_value, context_value
            )
        elif rule_value.comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_CONTAINED_WITHIN:
            return evaluate_contained_within(rule_value.comparison_value, context_value)
        elif rule_value.comparison_operator in (
            ComparisonOperator.COMPARISON_OPERATOR_STARTS_WITH,
            ComparisonOperator.COMPARISON_OPERATOR_ENDS_WITH,
            ComparisonOperator.COMPARISON_OPERATOR_CONTAINS,
        ):
            return evaluate_string_comparator(
                rule_value.comparison_operator, rule_value.comparison_value, context_value
            )
        else:
            raise EvaluationError("Unknown comparison operator")
    elif rule_type == "call_expression":
        if rule_value.WhichOneof("function") == "bucket":
            return evaluate_bucket(rule_value.bucket, namespace, config_name, context)
        else:
            raise EvaluationError("Unknown CallExpression function type")
    else:
        raise EvaluationError("Unknown rule type")


def evaluate_equals(rule_value: Value, context_value: LekkoValue) -> bool:
    rule_kind = rule_value.WhichOneof("kind") or ""
    context_kind = context_value.WhichOneof("kind") or ""
    if rule_kind not in ["bool_value", "string_value", "number_value"]:
        raise EvaluationError("Unsupported rule type for equals operator")

    if rule_kind == "number_value":
        if context_kind not in ["double_value", "int_value"]:
            raise EvaluationError("Type mismatch in equals operator rule")
    elif rule_kind != context_kind:
        raise EvaluationError("Type mismatch in equals operator rule")

    return getattr(rule_value, rule_kind) == getattr(context_value, context_kind)


def evaluate_string_comparator(
    comparison_operator: ComparisonOperator, rule_value: Value, context_value: LekkoValue
) -> bool:
    rule_str = get_string(rule_value)
    context_str = get_string(context_value)

    if comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_STARTS_WITH:
        return context_str.startswith(rule_str)
    elif comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_ENDS_WITH:
        return context_str.endswith(rule_str)
    elif comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_CONTAINS:
        return rule_str in context_str
    else:
        raise EvaluationError("Unknown string comparison operator")


def get_string(value: Value | LekkoValue) -> str:
    if not value:
        raise EvaluationError("String Value is undefined")

    if value.WhichOneof("kind") == "string_value":
        return value.string_value
    else:
        raise EvaluationError("get_string called with non-string Value")


def evaluate_number_comparator(
    comparison_operator: ComparisonOperator, rule_value: Value, context_value: LekkoValue
) -> bool:
    rule_num = get_number(rule_value)
    context_num = get_number(context_value)

    if comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN:
        return context_num < rule_num
    elif comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_LESS_THAN_OR_EQUALS:
        return context_num <= rule_num
    elif comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN:
        return context_num > rule_num
    elif comparison_operator == ComparisonOperator.COMPARISON_OPERATOR_GREATER_THAN_OR_EQUALS:
        return context_num >= rule_num
    else:
        raise EvaluationError("Unknown numerical comparison operator")


def get_number(value: Value | LekkoValue) -> float:
    value_kind = value.WhichOneof("kind")
    if value_kind in ["number_value", "int_value", "double_value"]:
        return float(getattr(value, value_kind))
    else:
        raise EvaluationError("get_number caled with non-numeric Value")


def evaluate_contained_within(rule_value: Value, context_value: LekkoValue) -> bool:
    if rule_value.WhichOneof("kind") != "list_value":
        raise EvaluationError("Contained within operator must use a list value")

    # TODO: this will throw if there's a type mismatch, which means that all items in rule list must be of same type
    # This is consistent with other language SDKs, but we should consider just returning False on type mismatch
    return any(evaluate_equals(list_elem_val, context_value) for list_elem_val in rule_value.list_value.values)


def evaluate_bucket(bucket_f: CallExpression.Bucket, namespace: str, config_name: str, context: ClientContext) -> bool:
    ctx_key = bucket_f.context_key
    value = context.get(ctx_key) if context else None
    if not value:
        # If key is missing in context map, evaluate to false - move to next rule
        return False

    bytes_buffer: bytes = b""

    value_kind = value.WhichOneof("kind")
    if not value_kind:
        return False

    value_val = getattr(value, value_kind)
    if value_kind == "string_value":
        bytes_buffer = bytes(value_val, "utf-8")
    elif value_kind == "int_value":
        bytes_buffer = value_val.to_bytes(8, byteorder="big")
    elif value_kind == "double_value":
        bytes_buffer = struct.pack(">d", value_val)
    else:
        raise EvaluationError("Unsupported value type for bucket")

    bytes_frags = [bytes(namespace, "utf-8"), bytes(config_name, "utf-8"), bytes(ctx_key, "utf-8"), bytes_buffer]
    result = xxh32(b"".join(bytes_frags), 0).intdigest()
    return result % 100000 <= bucket_f.threshold



from dataclasses import dataclass
from typing import Dict, Optional

from lekko_client.gen.lekko.client.v1beta1.configuration_service_pb2 import Value
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import Feature

ClientContext = Optional[Dict[str, Value]]


@dataclass
class FeatureData:
    config_sha: str
    feature: Feature



import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Type, TypeVar

from google.protobuf.message import Message as ProtoMessage


class Client(ABC):
    ProtoType = TypeVar("ProtoType", bound=ProtoMessage)

    def __init__(
        self,
        owner_name: str,
        repo_name: str,
        api_key: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
    ):
        self.owner_name = owner_name
        self.repo_name = repo_name
        self.context = context or {}

        self.api_key = api_key or os.environ.get("LEKKO_API_KEY")

    @abstractmethod
    def get_bool(self, namespace: str, key: str, context: Dict[str, Any]) -> bool:
        ...

    @abstractmethod
    def get_int(self, namespace: str, key: str, context: Dict[str, Any]) -> int:
        ...

    @abstractmethod
    def get_float(self, namespace: str, key: str, context: Dict[str, Any]) -> float:
        ...

    @abstractmethod
    def get_string(self, namespace: str, key: str, context: Dict[str, Any]) -> str:
        ...

    @abstractmethod
    def get_json(self, namespace: str, key: str, context: Dict[str, Any]) -> dict:
        ...

    @abstractmethod
    def get_proto(
        self,
        namsespace: str,
        key: str,
        context: Dict[str, Any],
    ) -> ProtoMessage:
        ...

    @abstractmethod
    def get_proto_by_type(
        self,
        namsespace: str,
        key: str,
        context: Dict[str, Any],
        proto_message_type: Type[ProtoType],
    ) -> ProtoType:
        ...

    @abstractmethod
    def close(self):
        ...



from typing import Dict

from lekko_client.exceptions import FeatureNotFound, NamespaceNotFound
from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    GetRepositoryContentsResponse,
)
from lekko_client.models import FeatureData
from lekko_client.stores.store import Store


class MemoryStore(Store):
    def __init__(self) -> None:
        super().__init__()
        self.configs: Dict[str, Dict[str, FeatureData]] = {}

    def get(self, namespace: str, config_key: str) -> FeatureData:
        namespace_map = self.configs.get(namespace)
        if not namespace_map:
            raise NamespaceNotFound(f"Namespace {namespace} not found")
        result = namespace_map.get(config_key)
        if not result:
            raise FeatureNotFound(f"Feature {config_key} not found in namespace {namespace}")
        return result

    def load_impl(self, contents: GetRepositoryContentsResponse) -> bool:
        new_configs = {}
        for ns in contents.namespaces:
            namespace_map = {}
            for cfg in ns.features:
                if cfg.feature:
                    namespace_map[cfg.name] = FeatureData(cfg.sha, cfg.feature)
            new_configs[ns.name] = namespace_map
        self.configs = new_configs
        return True



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 Lekko Technologies, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import lekko_client.gen.lekko.feature.v1beta1.feature_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class RepositoryKey(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OWNER_NAME_FIELD_NUMBER: builtins.int
    REPO_NAME_FIELD_NUMBER: builtins.int
    owner_name: builtins.str
    repo_name: builtins.str
    def __init__(
        self,
        *,
        owner_name: builtins.str = ...,
        repo_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["owner_name", b"owner_name", "repo_name", b"repo_name"]) -> None: ...

global___RepositoryKey = RepositoryKey

@typing_extensions.final
class GetRepositoryVersionRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPO_KEY_FIELD_NUMBER: builtins.int
    SESSION_KEY_FIELD_NUMBER: builtins.int
    @property
    def repo_key(self) -> global___RepositoryKey: ...
    session_key: builtins.str
    def __init__(
        self,
        *,
        repo_key: global___RepositoryKey | None = ...,
        session_key: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["repo_key", b"repo_key"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["repo_key", b"repo_key", "session_key", b"session_key"]) -> None: ...

global___GetRepositoryVersionRequest = GetRepositoryVersionRequest

@typing_extensions.final
class GetRepositoryVersionResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMIT_SHA_FIELD_NUMBER: builtins.int
    commit_sha: builtins.str
    def __init__(
        self,
        *,
        commit_sha: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["commit_sha", b"commit_sha"]) -> None: ...

global___GetRepositoryVersionResponse = GetRepositoryVersionResponse

@typing_extensions.final
class GetRepositoryContentsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPO_KEY_FIELD_NUMBER: builtins.int
    NAMESPACE_NAME_FIELD_NUMBER: builtins.int
    FEATURE_NAME_FIELD_NUMBER: builtins.int
    SESSION_KEY_FIELD_NUMBER: builtins.int
    @property
    def repo_key(self) -> global___RepositoryKey: ...
    namespace_name: builtins.str
    """optional namespace_name to filter responses by"""
    feature_name: builtins.str
    """optional feature_name to filter responses by"""
    session_key: builtins.str
    def __init__(
        self,
        *,
        repo_key: global___RepositoryKey | None = ...,
        namespace_name: builtins.str = ...,
        feature_name: builtins.str = ...,
        session_key: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["repo_key", b"repo_key"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["feature_name", b"feature_name", "namespace_name", b"namespace_name", "repo_key", b"repo_key", "session_key", b"session_key"]) -> None: ...

global___GetRepositoryContentsRequest = GetRepositoryContentsRequest

@typing_extensions.final
class GetRepositoryContentsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMIT_SHA_FIELD_NUMBER: builtins.int
    NAMESPACES_FIELD_NUMBER: builtins.int
    commit_sha: builtins.str
    @property
    def namespaces(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Namespace]: ...
    def __init__(
        self,
        *,
        commit_sha: builtins.str = ...,
        namespaces: collections.abc.Iterable[global___Namespace] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["commit_sha", b"commit_sha", "namespaces", b"namespaces"]) -> None: ...

global___GetRepositoryContentsResponse = GetRepositoryContentsResponse

@typing_extensions.final
class Namespace(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Feature]: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        features: collections.abc.Iterable[global___Feature] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["features", b"features", "name", b"name"]) -> None: ...

global___Namespace = Namespace

@typing_extensions.final
class Feature(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    SHA_FIELD_NUMBER: builtins.int
    FEATURE_FIELD_NUMBER: builtins.int
    name: builtins.str
    sha: builtins.str
    """The sha of the protobuf binary according to git."""
    @property
    def feature(self) -> lekko_client.gen.lekko.feature.v1beta1.feature_pb2.Feature: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        sha: builtins.str = ...,
        feature: lekko_client.gen.lekko.feature.v1beta1.feature_pb2.Feature | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["feature", b"feature"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["feature", b"feature", "name", b"name", "sha", b"sha"]) -> None: ...

global___Feature = Feature

@typing_extensions.final
class ContextKey(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    key: builtins.str
    type: builtins.str
    def __init__(
        self,
        *,
        key: builtins.str = ...,
        type: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "type", b"type"]) -> None: ...

global___ContextKey = ContextKey

@typing_extensions.final
class FlagEvaluationEvent(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPO_KEY_FIELD_NUMBER: builtins.int
    COMMIT_SHA_FIELD_NUMBER: builtins.int
    FEATURE_SHA_FIELD_NUMBER: builtins.int
    NAMESPACE_NAME_FIELD_NUMBER: builtins.int
    FEATURE_NAME_FIELD_NUMBER: builtins.int
    CONTEXT_KEYS_FIELD_NUMBER: builtins.int
    RESULT_PATH_FIELD_NUMBER: builtins.int
    CLIENT_EVENT_TIME_FIELD_NUMBER: builtins.int
    @property
    def repo_key(self) -> global___RepositoryKey: ...
    commit_sha: builtins.str
    feature_sha: builtins.str
    namespace_name: builtins.str
    feature_name: builtins.str
    @property
    def context_keys(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ContextKey]:
        """A list of context keys (not values) that were provided at runtime."""
    @property
    def result_path(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """The node in the tree that contained the final return value of the feature."""
    @property
    def client_event_time(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    def __init__(
        self,
        *,
        repo_key: global___RepositoryKey | None = ...,
        commit_sha: builtins.str = ...,
        feature_sha: builtins.str = ...,
        namespace_name: builtins.str = ...,
        feature_name: builtins.str = ...,
        context_keys: collections.abc.Iterable[global___ContextKey] | None = ...,
        result_path: collections.abc.Iterable[builtins.int] | None = ...,
        client_event_time: google.protobuf.timestamp_pb2.Timestamp | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["client_event_time", b"client_event_time", "repo_key", b"repo_key"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["client_event_time", b"client_event_time", "commit_sha", b"commit_sha", "context_keys", b"context_keys", "feature_name", b"feature_name", "feature_sha", b"feature_sha", "namespace_name", b"namespace_name", "repo_key", b"repo_key", "result_path", b"result_path"]) -> None: ...

global___FlagEvaluationEvent = FlagEvaluationEvent

@typing_extensions.final
class SendFlagEvaluationMetricsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVENTS_FIELD_NUMBER: builtins.int
    SESSION_KEY_FIELD_NUMBER: builtins.int
    @property
    def events(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___FlagEvaluationEvent]: ...
    session_key: builtins.str
    def __init__(
        self,
        *,
        events: collections.abc.Iterable[global___FlagEvaluationEvent] | None = ...,
        session_key: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["events", b"events", "session_key", b"session_key"]) -> None: ...

global___SendFlagEvaluationMetricsRequest = SendFlagEvaluationMetricsRequest

@typing_extensions.final
class SendFlagEvaluationMetricsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___SendFlagEvaluationMetricsResponse = SendFlagEvaluationMetricsResponse

@typing_extensions.final
class RegisterClientRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPO_KEY_FIELD_NUMBER: builtins.int
    NAMESPACE_LIST_FIELD_NUMBER: builtins.int
    INITIAL_BOOTSTRAP_SHA_FIELD_NUMBER: builtins.int
    SIDECAR_VERSION_FIELD_NUMBER: builtins.int
    @property
    def repo_key(self) -> global___RepositoryKey: ...
    @property
    def namespace_list(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The namespaces to register within the repo. If empty,
        all namespaces will be registered.
        """
    initial_bootstrap_sha: builtins.str
    """If the client was initialized from a git bootstrap,
    the commit sha is provided. If there was no bootstrap, this
    can be an empty string.
    """
    sidecar_version: builtins.str
    """If the client is a lekko sidecar, provide the semver version,
    or if not available, the sha of the sidecar.
    """
    def __init__(
        self,
        *,
        repo_key: global___RepositoryKey | None = ...,
        namespace_list: collections.abc.Iterable[builtins.str] | None = ...,
        initial_bootstrap_sha: builtins.str = ...,
        sidecar_version: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["repo_key", b"repo_key"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["initial_bootstrap_sha", b"initial_bootstrap_sha", "namespace_list", b"namespace_list", "repo_key", b"repo_key", "sidecar_version", b"sidecar_version"]) -> None: ...

global___RegisterClientRequest = RegisterClientRequest

@typing_extensions.final
class RegisterClientResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SESSION_KEY_FIELD_NUMBER: builtins.int
    session_key: builtins.str
    """TODO make this field 1 if we rewrite the API."""
    def __init__(
        self,
        *,
        session_key: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["session_key", b"session_key"]) -> None: ...

global___RegisterClientResponse = RegisterClientResponse

@typing_extensions.final
class DeregisterClientRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SESSION_KEY_FIELD_NUMBER: builtins.int
    session_key: builtins.str
    def __init__(
        self,
        *,
        session_key: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["session_key", b"session_key"]) -> None: ...

global___DeregisterClientRequest = DeregisterClientRequest

@typing_extensions.final
class DeregisterClientResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___DeregisterClientResponse = DeregisterClientResponse

@typing_extensions.final
class GetDeveloperAccessTokenRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___GetDeveloperAccessTokenRequest = GetDeveloperAccessTokenRequest

@typing_extensions.final
class GetDeveloperAccessTokenResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    token: builtins.str
    """github access token"""
    def __init__(
        self,
        *,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["token", b"token"]) -> None: ...

global___GetDeveloperAccessTokenResponse = GetDeveloperAccessTokenResponse



from abc import ABC, abstractmethod
from hashlib import sha256

from lekko_client.gen.lekko.backend.v1beta1.distribution_service_pb2 import (
    GetRepositoryContentsResponse,
)
from lekko_client.models import FeatureData


class Store(ABC):
    def __init__(self):
        self._commit_sha = ""
        self._content_hash = ""

    @abstractmethod
    def get(self, namespace: str, config_key: str) -> FeatureData:
        ...

    def load(self, contents: GetRepositoryContentsResponse) -> bool:
        if not contents:
            return False

        contents = self.sort_contents(contents)
        content_hash = self.hash_contents(contents)

        if not self.should_update(contents, content_hash):
            return False

        if not self.load_impl(contents):
            return False

        self._commit_sha = contents.commit_sha
        self._content_hash = content_hash
        return True

    @abstractmethod
    def load_impl(self, contents: GetRepositoryContentsResponse) -> bool:
        ...

    @property
    def commit_sha(self) -> str:
        return self._commit_sha

    @property
    def content_hash(self) -> str:
        return self._content_hash

    def should_update(self, contents: GetRepositoryContentsResponse, content_hash: str):
        ret = contents.commit_sha != self.commit_sha or content_hash != self.content_hash
        return ret

    @classmethod
    def sort_contents(cls, contents: GetRepositoryContentsResponse):
        for ns in contents.namespaces:
            ns.features.sort(key=lambda cfg: cfg.name)
        contents.namespaces.sort(key=lambda ns: ns.name)
        return contents

    @classmethod
    def hash_contents(cls, contents: GetRepositoryContentsResponse):
        return sha256(contents.SerializeToString()).hexdigest()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 Lekko Technologies, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _ComparisonOperator:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ComparisonOperatorEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ComparisonOperator.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    COMPARISON_OPERATOR_UNSPECIFIED: _ComparisonOperator.ValueType  # 0
    COMPARISON_OPERATOR_EQUALS: _ComparisonOperator.ValueType  # 1
    """== only applies to number, string and bool values."""
    COMPARISON_OPERATOR_LESS_THAN: _ComparisonOperator.ValueType  # 2
    """> < >= <= only applies to number values."""
    COMPARISON_OPERATOR_LESS_THAN_OR_EQUALS: _ComparisonOperator.ValueType  # 3
    COMPARISON_OPERATOR_GREATER_THAN: _ComparisonOperator.ValueType  # 4
    COMPARISON_OPERATOR_GREATER_THAN_OR_EQUALS: _ComparisonOperator.ValueType  # 5
    COMPARISON_OPERATOR_CONTAINED_WITHIN: _ComparisonOperator.ValueType  # 6
    """Contained within only applies to list values. Elements
    of the list must be primitive (i.e. number, string or bool)
    """
    COMPARISON_OPERATOR_STARTS_WITH: _ComparisonOperator.ValueType  # 7
    """Starts with and ends with only apply to string values."""
    COMPARISON_OPERATOR_ENDS_WITH: _ComparisonOperator.ValueType  # 8
    COMPARISON_OPERATOR_CONTAINS: _ComparisonOperator.ValueType  # 9
    """Contains only applies to string values, and for now is strict equality.
    If we support things like regex or case insensitive matches, they will
    be separate operators.
    """
    COMPARISON_OPERATOR_PRESENT: _ComparisonOperator.ValueType  # 10
    """Present is the only operator that doesn't require a comparison value."""

class ComparisonOperator(_ComparisonOperator, metaclass=_ComparisonOperatorEnumTypeWrapper): ...

COMPARISON_OPERATOR_UNSPECIFIED: ComparisonOperator.ValueType  # 0
COMPARISON_OPERATOR_EQUALS: ComparisonOperator.ValueType  # 1
"""== only applies to number, string and bool values."""
COMPARISON_OPERATOR_LESS_THAN: ComparisonOperator.ValueType  # 2
"""> < >= <= only applies to number values."""
COMPARISON_OPERATOR_LESS_THAN_OR_EQUALS: ComparisonOperator.ValueType  # 3
COMPARISON_OPERATOR_GREATER_THAN: ComparisonOperator.ValueType  # 4
COMPARISON_OPERATOR_GREATER_THAN_OR_EQUALS: ComparisonOperator.ValueType  # 5
COMPARISON_OPERATOR_CONTAINED_WITHIN: ComparisonOperator.ValueType  # 6
"""Contained within only applies to list values. Elements
of the list must be primitive (i.e. number, string or bool)
"""
COMPARISON_OPERATOR_STARTS_WITH: ComparisonOperator.ValueType  # 7
"""Starts with and ends with only apply to string values."""
COMPARISON_OPERATOR_ENDS_WITH: ComparisonOperator.ValueType  # 8
COMPARISON_OPERATOR_CONTAINS: ComparisonOperator.ValueType  # 9
"""Contains only applies to string values, and for now is strict equality.
If we support things like regex or case insensitive matches, they will
be separate operators.
"""
COMPARISON_OPERATOR_PRESENT: ComparisonOperator.ValueType  # 10
"""Present is the only operator that doesn't require a comparison value."""
global___ComparisonOperator = ComparisonOperator

class _LogicalOperator:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _LogicalOperatorEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_LogicalOperator.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    LOGICAL_OPERATOR_UNSPECIFIED: _LogicalOperator.ValueType  # 0
    LOGICAL_OPERATOR_AND: _LogicalOperator.ValueType  # 1
    LOGICAL_OPERATOR_OR: _LogicalOperator.ValueType  # 2

class LogicalOperator(_LogicalOperator, metaclass=_LogicalOperatorEnumTypeWrapper): ...

LOGICAL_OPERATOR_UNSPECIFIED: LogicalOperator.ValueType  # 0
LOGICAL_OPERATOR_AND: LogicalOperator.ValueType  # 1
LOGICAL_OPERATOR_OR: LogicalOperator.ValueType  # 2
global___LogicalOperator = LogicalOperator

@typing_extensions.final
class Rule(google.protobuf.message.Message):
    """A Rule is a top level object that recursively defines an AST represented
    by ruleslang. A rule is always one of 4 things:
    1. Atom -> This is a leaf node in the tree that returns true or false
    2. Not -> This negates the result of the underlying Rule.
    3. LogicalExpression -> This rule links two rules through an "and" or an "or".
    4. BoolConst -> true or false. This will be used for higher level short-circuits.
    Parenthases and other logical constructs can all be represented by the correct
    construction of this rule tree.

    !(A && B && C) || D can be represented by LogExp ( Not ( LogExp ( LogExp ( Atom(A) && Atom(B) ) && Atom (C))) || Atom(D))
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ATOM_FIELD_NUMBER: builtins.int
    NOT_FIELD_NUMBER: builtins.int
    LOGICAL_EXPRESSION_FIELD_NUMBER: builtins.int
    BOOL_CONST_FIELD_NUMBER: builtins.int
    @property
    def atom(self) -> global___Atom: ...
    @property
    def logical_expression(self) -> global___LogicalExpression: ...
    bool_const: builtins.bool
    def __init__(
        self,
        *,
        atom: global___Atom | None = ...,
        logical_expression: global___LogicalExpression | None = ...,
        bool_const: builtins.bool = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["atom", b"atom", "bool_const", b"bool_const", "logical_expression", b"logical_expression", "not", b"not", "rule", b"rule"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["atom", b"atom", "bool_const", b"bool_const", "logical_expression", b"logical_expression", "not", b"not", "rule", b"rule"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["rule", b"rule"]) -> typing_extensions.Literal["atom", "not", "logical_expression", "bool_const"] | None: ...

global___Rule = Rule

@typing_extensions.final
class LogicalExpression(google.protobuf.message.Message):
    """LogicalExpression operator applies a logical operator like "and" or "or" to two rules.
    They are evaluated in the order expressed by the field numbers and field names, with "first_rule" first.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FIRST_RULE_FIELD_NUMBER: builtins.int
    SECOND_RULE_FIELD_NUMBER: builtins.int
    LOGICAL_OPERATOR_FIELD_NUMBER: builtins.int
    @property
    def first_rule(self) -> global___Rule: ...
    @property
    def second_rule(self) -> global___Rule: ...
    logical_operator: global___LogicalOperator.ValueType
    def __init__(
        self,
        *,
        first_rule: global___Rule | None = ...,
        second_rule: global___Rule | None = ...,
        logical_operator: global___LogicalOperator.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["first_rule", b"first_rule", "second_rule", b"second_rule"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["first_rule", b"first_rule", "logical_operator", b"logical_operator", "second_rule", b"second_rule"]) -> None: ...

global___LogicalExpression = LogicalExpression

@typing_extensions.final
class Atom(google.protobuf.message.Message):
    """An atom is a fragment of ruleslang that can result in a true or false.
    An atom always has a comparison operator and a context key, and can optionally
    have a comparison value.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTEXT_KEY_FIELD_NUMBER: builtins.int
    COMPARISON_VALUE_FIELD_NUMBER: builtins.int
    COMPARISON_OPERATOR_FIELD_NUMBER: builtins.int
    context_key: builtins.str
    @property
    def comparison_value(self) -> google.protobuf.struct_pb2.Value:
        """For the "PRESENT" operator, the comparison value should be null."""
    comparison_operator: global___ComparisonOperator.ValueType
    """For operators, context is on the left, comparison value on the right."""
    def __init__(
        self,
        *,
        context_key: builtins.str = ...,
        comparison_value: google.protobuf.struct_pb2.Value | None = ...,
        comparison_operator: global___ComparisonOperator.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["comparison_value", b"comparison_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["comparison_operator", b"comparison_operator", "comparison_value", b"comparison_value", "context_key", b"context_key"]) -> None: ...

global___Atom = Atom



from lekko_client.clients.cached_backend_client import CachedBackendClient  # noqa
from lekko_client.clients.cached_git_client import CachedGitClient  # noqa
from lekko_client.clients.client import Client  # noqa
from lekko_client.clients.config_client import (  # noqa
    APIClient,
    ConfigServiceClient,
    SidecarClient,
)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from lekko_client.gen.lekko.backend.v1beta1 import distribution_service_pb2 as lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2


class DistributionServiceStub(object):
    """Initial implementation of a config distribution service. Clients should begin by
    calling the register RPC which returns a session key, which is used in all other RPCs.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetRepositoryVersion = channel.unary_unary(
                '/lekko.backend.v1beta1.DistributionService/GetRepositoryVersion',
                request_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryVersionRequest.SerializeToString,
                response_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryVersionResponse.FromString,
                )
        self.GetRepositoryContents = channel.unary_unary(
                '/lekko.backend.v1beta1.DistributionService/GetRepositoryContents',
                request_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryContentsRequest.SerializeToString,
                response_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryContentsResponse.FromString,
                )
        self.SendFlagEvaluationMetrics = channel.unary_unary(
                '/lekko.backend.v1beta1.DistributionService/SendFlagEvaluationMetrics',
                request_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.SendFlagEvaluationMetricsRequest.SerializeToString,
                response_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.SendFlagEvaluationMetricsResponse.FromString,
                )
        self.RegisterClient = channel.unary_unary(
                '/lekko.backend.v1beta1.DistributionService/RegisterClient',
                request_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.RegisterClientRequest.SerializeToString,
                response_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.RegisterClientResponse.FromString,
                )
        self.DeregisterClient = channel.unary_unary(
                '/lekko.backend.v1beta1.DistributionService/DeregisterClient',
                request_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.DeregisterClientRequest.SerializeToString,
                response_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.DeregisterClientResponse.FromString,
                )
        self.GetDeveloperAccessToken = channel.unary_unary(
                '/lekko.backend.v1beta1.DistributionService/GetDeveloperAccessToken',
                request_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetDeveloperAccessTokenRequest.SerializeToString,
                response_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetDeveloperAccessTokenResponse.FromString,
                )


class DistributionServiceServicer(object):
    """Initial implementation of a config distribution service. Clients should begin by
    calling the register RPC which returns a session key, which is used in all other RPCs.
    """

    def GetRepositoryVersion(self, request, context):
        """Returns the latest commit sha of the repository. The client is expected to poll this
        rpc to become aware of updates.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRepositoryContents(self, request, context):
        """Returns the entire state of the repository, including all feature flags.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendFlagEvaluationMetrics(self, request, context):
        """Sends metrics to the backend related to flag evaluation. This rpc can be used
        to batch metrics to lekko servers.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterClient(self, request, context):
        """Registers a client and returns a session key.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeregisterClient(self, request, context):
        """Deregisters a client using a session key.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDeveloperAccessToken(self, request, context):
        """Returns a github access token that provides access to the contents of
        some of Lekko's private repositories
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DistributionServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetRepositoryVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRepositoryVersion,
                    request_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryVersionRequest.FromString,
                    response_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryVersionResponse.SerializeToString,
            ),
            'GetRepositoryContents': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRepositoryContents,
                    request_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryContentsRequest.FromString,
                    response_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryContentsResponse.SerializeToString,
            ),
            'SendFlagEvaluationMetrics': grpc.unary_unary_rpc_method_handler(
                    servicer.SendFlagEvaluationMetrics,
                    request_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.SendFlagEvaluationMetricsRequest.FromString,
                    response_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.SendFlagEvaluationMetricsResponse.SerializeToString,
            ),
            'RegisterClient': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterClient,
                    request_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.RegisterClientRequest.FromString,
                    response_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.RegisterClientResponse.SerializeToString,
            ),
            'DeregisterClient': grpc.unary_unary_rpc_method_handler(
                    servicer.DeregisterClient,
                    request_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.DeregisterClientRequest.FromString,
                    response_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.DeregisterClientResponse.SerializeToString,
            ),
            'GetDeveloperAccessToken': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDeveloperAccessToken,
                    request_deserializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetDeveloperAccessTokenRequest.FromString,
                    response_serializer=lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetDeveloperAccessTokenResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'lekko.backend.v1beta1.DistributionService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class DistributionService(object):
    """Initial implementation of a config distribution service. Clients should begin by
    calling the register RPC which returns a session key, which is used in all other RPCs.
    """

    @staticmethod
    def GetRepositoryVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lekko.backend.v1beta1.DistributionService/GetRepositoryVersion',
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryVersionRequest.SerializeToString,
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryVersionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRepositoryContents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lekko.backend.v1beta1.DistributionService/GetRepositoryContents',
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryContentsRequest.SerializeToString,
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetRepositoryContentsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendFlagEvaluationMetrics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lekko.backend.v1beta1.DistributionService/SendFlagEvaluationMetrics',
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.SendFlagEvaluationMetricsRequest.SerializeToString,
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.SendFlagEvaluationMetricsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RegisterClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lekko.backend.v1beta1.DistributionService/RegisterClient',
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.RegisterClientRequest.SerializeToString,
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.RegisterClientResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeregisterClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lekko.backend.v1beta1.DistributionService/DeregisterClient',
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.DeregisterClientRequest.SerializeToString,
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.DeregisterClientResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetDeveloperAccessToken(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/lekko.backend.v1beta1.DistributionService/GetDeveloperAccessToken',
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetDeveloperAccessTokenRequest.SerializeToString,
            lekko_dot_backend_dot_v1beta1_dot_distribution__service__pb2.GetDeveloperAccessTokenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lekko/rules/v1beta3/rules.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1flekko/rules/v1beta3/rules.proto\x12\x13lekko.rules.v1beta3\x1a\x1cgoogle/protobuf/struct.proto\"\xb8\x02\n\x04Rule\x12/\n\x04\x61tom\x18\x01 \x01(\x0b\x32\x19.lekko.rules.v1beta3.AtomH\x00R\x04\x61tom\x12-\n\x03not\x18\x02 \x01(\x0b\x32\x19.lekko.rules.v1beta3.RuleH\x00R\x03not\x12W\n\x12logical_expression\x18\x03 \x01(\x0b\x32&.lekko.rules.v1beta3.LogicalExpressionH\x00R\x11logicalExpression\x12\x1f\n\nbool_const\x18\x04 \x01(\x08H\x00R\tboolConst\x12N\n\x0f\x63\x61ll_expression\x18\x05 \x01(\x0b\x32#.lekko.rules.v1beta3.CallExpressionH\x00R\x0e\x63\x61llExpressionB\x06\n\x04rule\"\x95\x01\n\x11LogicalExpression\x12/\n\x05rules\x18\x01 \x03(\x0b\x32\x19.lekko.rules.v1beta3.RuleR\x05rules\x12O\n\x10logical_operator\x18\x03 \x01(\x0e\x32$.lekko.rules.v1beta3.LogicalOperatorR\x0flogicalOperator\"\xc4\x01\n\x04\x41tom\x12\x1f\n\x0b\x63ontext_key\x18\x01 \x01(\tR\ncontextKey\x12\x41\n\x10\x63omparison_value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.ValueR\x0f\x63omparisonValue\x12X\n\x13\x63omparison_operator\x18\x03 \x01(\x0e\x32\'.lekko.rules.v1beta3.ComparisonOperatorR\x12\x63omparisonOperator\"\xab\x01\n\x0e\x43\x61llExpression\x12\x44\n\x06\x62ucket\x18\x01 \x01(\x0b\x32*.lekko.rules.v1beta3.CallExpression.BucketH\x00R\x06\x62ucket\x1aG\n\x06\x42ucket\x12\x1f\n\x0b\x63ontext_key\x18\x01 \x01(\tR\ncontextKey\x12\x1c\n\tthreshold\x18\x02 \x01(\rR\tthresholdB\n\n\x08\x66unction*\xd8\x03\n\x12\x43omparisonOperator\x12#\n\x1f\x43OMPARISON_OPERATOR_UNSPECIFIED\x10\x00\x12\x1e\n\x1a\x43OMPARISON_OPERATOR_EQUALS\x10\x01\x12!\n\x1d\x43OMPARISON_OPERATOR_LESS_THAN\x10\x02\x12+\n\'COMPARISON_OPERATOR_LESS_THAN_OR_EQUALS\x10\x03\x12$\n COMPARISON_OPERATOR_GREATER_THAN\x10\x04\x12.\n*COMPARISON_OPERATOR_GREATER_THAN_OR_EQUALS\x10\x05\x12(\n$COMPARISON_OPERATOR_CONTAINED_WITHIN\x10\x06\x12#\n\x1f\x43OMPARISON_OPERATOR_STARTS_WITH\x10\x07\x12!\n\x1d\x43OMPARISON_OPERATOR_ENDS_WITH\x10\x08\x12 \n\x1c\x43OMPARISON_OPERATOR_CONTAINS\x10\t\x12\x1f\n\x1b\x43OMPARISON_OPERATOR_PRESENT\x10\n\x12\"\n\x1e\x43OMPARISON_OPERATOR_NOT_EQUALS\x10\x0b*f\n\x0fLogicalOperator\x12 \n\x1cLOGICAL_OPERATOR_UNSPECIFIED\x10\x00\x12\x18\n\x14LOGICAL_OPERATOR_AND\x10\x01\x12\x17\n\x13LOGICAL_OPERATOR_OR\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lekko.rules.v1beta3.rules_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_COMPARISONOPERATOR']._serialized_start=927
  _globals['_COMPARISONOPERATOR']._serialized_end=1399
  _globals['_LOGICALOPERATOR']._serialized_start=1401
  _globals['_LOGICALOPERATOR']._serialized_end=1503
  _globals['_RULE']._serialized_start=87
  _globals['_RULE']._serialized_end=399
  _globals['_LOGICALEXPRESSION']._serialized_start=402
  _globals['_LOGICALEXPRESSION']._serialized_end=551
  _globals['_ATOM']._serialized_start=554
  _globals['_ATOM']._serialized_end=750
  _globals['_CALLEXPRESSION']._serialized_start=753
  _globals['_CALLEXPRESSION']._serialized_end=924
  _globals['_CALLEXPRESSION_BUCKET']._serialized_start=841
  _globals['_CALLEXPRESSION_BUCKET']._serialized_end=912
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lekko/feature/v1beta1/static.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from lekko_client.gen.lekko.feature.v1beta1 import feature_pb2 as lekko_dot_feature_dot_v1beta1_dot_feature__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"lekko/feature/v1beta1/static.proto\x12\x15lekko.feature.v1beta1\x1a#lekko/feature/v1beta1/feature.proto\"\x9c\x02\n\rStaticFeature\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x36\n\x04type\x18\x02 \x01(\x0e\x32\".lekko.feature.v1beta1.FeatureTypeR\x04type\x12@\n\x07imports\x18\x03 \x03(\x0b\x32&.lekko.feature.v1beta1.ImportStatementR\x07imports\x12>\n\x07\x66\x65\x61ture\x18\x04 \x01(\x0b\x32$.lekko.feature.v1beta1.FeatureStructR\x07\x66\x65\x61ture\x12?\n\x0b\x66\x65\x61ture_old\x18\x05 \x01(\x0b\x32\x1e.lekko.feature.v1beta1.FeatureR\nfeatureOld\"\xd5\x01\n\rFeatureStruct\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\x12\x39\n\x07\x64\x65\x66\x61ult\x18\x03 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarExprR\x07\x64\x65\x66\x61ult\x12\x32\n\x05rules\x18\x04 \x01(\x0b\x32\x1c.lekko.feature.v1beta1.RulesR\x05rules\"o\n\x05Rules\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12\x31\n\x05rules\x18\x02 \x03(\x0b\x32\x1b.lekko.feature.v1beta1.RuleR\x05rules\"\x90\x01\n\x04Rule\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12\x1c\n\tcondition\x18\x02 \x01(\tR\tcondition\x12\x35\n\x05value\x18\x03 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarExprR\x05value\"\xea\x01\n\x0fImportStatement\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12\x32\n\x03lhs\x18\x02 \x01(\x0b\x32 .lekko.feature.v1beta1.IdentExprR\x03lhs\x12\x1a\n\x08operator\x18\x03 \x01(\tR\x08operator\x12\x1d\n\nline_break\x18\x04 \x01(\x08R\tlineBreak\x12\x33\n\x03rhs\x18\x05 \x01(\x0b\x32!.lekko.feature.v1beta1.ImportExprR\x03rhs\"\x87\x01\n\nImportExpr\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12\x30\n\x03\x64ot\x18\x02 \x01(\x0b\x32\x1e.lekko.feature.v1beta1.DotExprR\x03\x64ot\x12\x12\n\x04\x61rgs\x18\x03 \x03(\tR\x04\x61rgs\"`\n\x07\x44otExpr\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12\x0c\n\x01x\x18\x02 \x01(\tR\x01x\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name\"_\n\x08StarExpr\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12\x1e\n\nexpression\x18\x02 \x01(\tR\nexpression\"V\n\tIdentExpr\x12\x33\n\x04meta\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.StarMetaR\x04meta\x12\x14\n\x05token\x18\x02 \x01(\tR\x05token\"e\n\x08StarMeta\x12;\n\x08\x63omments\x18\x01 \x01(\x0b\x32\x1f.lekko.feature.v1beta1.CommentsR\x08\x63omments\x12\x1c\n\tmultiline\x18\x02 \x01(\x08R\tmultiline\"\xb0\x01\n\x08\x43omments\x12\x36\n\x06\x62\x65\x66ore\x18\x01 \x03(\x0b\x32\x1e.lekko.feature.v1beta1.CommentR\x06\x62\x65\x66ore\x12\x36\n\x06suffix\x18\x02 \x03(\x0b\x32\x1e.lekko.feature.v1beta1.CommentR\x06suffix\x12\x34\n\x05\x61\x66ter\x18\x03 \x03(\x0b\x32\x1e.lekko.feature.v1beta1.CommentR\x05\x61\x66ter\"\x1f\n\x07\x43omment\x12\x14\n\x05token\x18\x01 \x01(\tR\x05tokenb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lekko.feature.v1beta1.static_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_STATICFEATURE']._serialized_start=99
  _globals['_STATICFEATURE']._serialized_end=383
  _globals['_FEATURESTRUCT']._serialized_start=386
  _globals['_FEATURESTRUCT']._serialized_end=599
  _globals['_RULES']._serialized_start=601
  _globals['_RULES']._serialized_end=712
  _globals['_RULE']._serialized_start=715
  _globals['_RULE']._serialized_end=859
  _globals['_IMPORTSTATEMENT']._serialized_start=862
  _globals['_IMPORTSTATEMENT']._serialized_end=1096
  _globals['_IMPORTEXPR']._serialized_start=1099
  _globals['_IMPORTEXPR']._serialized_end=1234
  _globals['_DOTEXPR']._serialized_start=1236
  _globals['_DOTEXPR']._serialized_end=1332
  _globals['_STAREXPR']._serialized_start=1334
  _globals['_STAREXPR']._serialized_end=1429
  _globals['_IDENTEXPR']._serialized_start=1431
  _globals['_IDENTEXPR']._serialized_end=1517
  _globals['_STARMETA']._serialized_start=1519
  _globals['_STARMETA']._serialized_end=1620
  _globals['_COMMENTS']._serialized_start=1623
  _globals['_COMMENTS']._serialized_end=1799
  _globals['_COMMENT']._serialized_start=1801
  _globals['_COMMENT']._serialized_end=1832
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lekko/feature/v1beta1/feature.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from lekko_client.gen.lekko.rules.v1beta2 import rules_pb2 as lekko_dot_rules_dot_v1beta2_dot_rules__pb2
from lekko_client.gen.lekko.rules.v1beta3 import rules_pb2 as lekko_dot_rules_dot_v1beta3_dot_rules__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#lekko/feature/v1beta1/feature.proto\x12\x15lekko.feature.v1beta1\x1a\x19google/protobuf/any.proto\x1a\x1flekko/rules/v1beta2/rules.proto\x1a\x1flekko/rules/v1beta3/rules.proto\"\xa6\x01\n\x07\x46\x65\x61ture\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\x12/\n\x04tree\x18\x03 \x01(\x0b\x32\x1b.lekko.feature.v1beta1.TreeR\x04tree\x12\x36\n\x04type\x18\x04 \x01(\x0e\x32\".lekko.feature.v1beta1.FeatureTypeR\x04type\"\xb8\x01\n\x04Tree\x12.\n\x07\x64\x65\x66\x61ult\x18\x01 \x01(\x0b\x32\x14.google.protobuf.AnyR\x07\x64\x65\x66\x61ult\x12\x43\n\x0b\x63onstraints\x18\x02 \x03(\x0b\x32!.lekko.feature.v1beta1.ConstraintR\x0b\x63onstraints\x12;\n\x0b\x64\x65\x66\x61ult_new\x18\x03 \x01(\x0b\x32\x1a.lekko.feature.v1beta1.AnyR\ndefaultNew\"\xc1\x02\n\nConstraint\x12\x12\n\x04rule\x18\x01 \x01(\tR\x04rule\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.AnyR\x05value\x12\x43\n\x0b\x63onstraints\x18\x03 \x03(\x0b\x32!.lekko.feature.v1beta1.ConstraintR\x0b\x63onstraints\x12\x38\n\x08rule_ast\x18\x04 \x01(\x0b\x32\x19.lekko.rules.v1beta2.RuleB\x02\x18\x01R\x07ruleAst\x12;\n\x0crule_ast_new\x18\x05 \x01(\x0b\x32\x19.lekko.rules.v1beta3.RuleR\nruleAstNew\x12\x37\n\tvalue_new\x18\x06 \x01(\x0b\x32\x1a.lekko.feature.v1beta1.AnyR\x08valueNew\"6\n\x03\x41ny\x12\x19\n\x08type_url\x18\x01 \x01(\tR\x07typeUrl\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value*\xb8\x01\n\x0b\x46\x65\x61tureType\x12\x1c\n\x18\x46\x45\x41TURE_TYPE_UNSPECIFIED\x10\x00\x12\x15\n\x11\x46\x45\x41TURE_TYPE_BOOL\x10\x01\x12\x14\n\x10\x46\x45\x41TURE_TYPE_INT\x10\x02\x12\x16\n\x12\x46\x45\x41TURE_TYPE_FLOAT\x10\x03\x12\x17\n\x13\x46\x45\x41TURE_TYPE_STRING\x10\x04\x12\x15\n\x11\x46\x45\x41TURE_TYPE_JSON\x10\x05\x12\x16\n\x12\x46\x45\x41TURE_TYPE_PROTO\x10\x06\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lekko.feature.v1beta1.feature_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _CONSTRAINT.fields_by_name['rule_ast']._options = None
  _CONSTRAINT.fields_by_name['rule_ast']._serialized_options = b'\030\001'
  _globals['_FEATURETYPE']._serialized_start=892
  _globals['_FEATURETYPE']._serialized_end=1076
  _globals['_FEATURE']._serialized_start=156
  _globals['_FEATURE']._serialized_end=322
  _globals['_TREE']._serialized_start=325
  _globals['_TREE']._serialized_end=509
  _globals['_CONSTRAINT']._serialized_start=512
  _globals['_CONSTRAINT']._serialized_end=833
  _globals['_ANY']._serialized_start=835
  _globals['_ANY']._serialized_end=889
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass
from typing import List, Optional

from google.protobuf.any_pb2 import Any as ProtoAny

from lekko_client.evaluation.rules import evaluate_rule
from lekko_client.exceptions import EvaluationError
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import Any as LekkoAny
from lekko_client.gen.lekko.feature.v1beta1.feature_pb2 import Constraint, Feature
from lekko_client.models import ClientContext


@dataclass
class EvaluationResult:
    value: ProtoAny
    # Stores the path of the tree node that returned the final value
    # after successful evaluation.
    path: List[int]


@dataclass
class TraverseResult:
    value: Optional[ProtoAny]
    passes: bool
    path: List[int]


def evaluate(config: Feature, namespace: str, context: ClientContext = None) -> EvaluationResult:
    if not config.HasField("tree"):
        raise EvaluationError("Unable to evaluate feature: config tree is empty")

    for i, constraint in enumerate(config.tree.constraints):
        child_result = traverse(constraint, namespace, config.key, context)
        if child_result.passes:
            if child_result.value:
                return EvaluationResult(value=child_result.value, path=[i, *child_result.path])
            break
    return EvaluationResult(value=_get_any(config.tree.default, config.tree.default_new), path=[])


def traverse(override: Constraint, namespace: str, config_name: str, context: ClientContext = None) -> TraverseResult:
    if not override:
        return TraverseResult(None, False, [])
    passes = evaluate_rule(override.rule_ast_new, namespace, config_name, context)
    if not passes:
        return TraverseResult(None, False, [])

    for i, constraint in enumerate(override.constraints):
        child_result = traverse(constraint, namespace, config_name, context)
        if child_result.passes:
            if child_result.value:
                return TraverseResult(child_result.value, True, [i, *child_result.path])
            break
    return TraverseResult(_get_any(override.value, override.value_new), True, [])


def _get_any(val: Optional[ProtoAny], val_new: Optional[LekkoAny]) -> ProtoAny:
    if val_new and val_new.type_url:
        return ProtoAny(type_url=val_new.type_url, value=val_new.value)
    if val:
        return val
    raise EvaluationError("Constraint or default value is empty")



"""Lekko Python SDK Client"""

import functools
import logging
from dataclasses import dataclass
from threading import RLock
from typing import Any, Dict, Optional, Type

from google.protobuf.message import Message as ProtoMessage

from lekko_client import exceptions
from lekko_client.clients import (
    APIClient,
    CachedBackendClient,
    CachedGitClient,
    Client,
    SidecarClient,
)
from lekko_client.constants import LEKKO_API_URL, LEKKO_SIDECAR_URL  # noqa
from lekko_client.stores import MemoryStore

__version__ = "0.1.5rc1"

logging.getLogger(__name__).addHandler(logging.NullHandler())

__client: Optional[Client] = None
__client_lock = RLock()


@dataclass(kw_only=True)
class Config:
    owner_name: str
    repo_name: str
    api_key: Optional[str] = None
    context: Optional[Dict[str, Any]] = None
    lekko_uri: str = LEKKO_API_URL


@dataclass(kw_only=True)
class SidecarConfig(Config):
    lekko_uri = LEKKO_SIDECAR_URL


@dataclass(kw_only=True)
class APIConfig(Config):
    pass


@dataclass(kw_only=True)
class CachedServerConfig(Config):
    pass


@dataclass(kw_only=True)
class CachedGitConfig(Config):
    git_repo_path: str


def initialize(config: Config) -> Client:
    global __client
    with __client_lock:
        if __client:
            __client.close()
        match config:
            case APIConfig():
                __client = APIClient(
                    owner_name=config.owner_name,
                    repo_name=config.repo_name,
                    api_key=config.api_key,
                    context=config.context,
                )
            case SidecarConfig():
                __client = SidecarClient(
                    owner_name=config.owner_name,
                    repo_name=config.repo_name,
                    api_key=config.api_key,
                    context=config.context,
                )
            case CachedGitConfig():
                __client = CachedGitClient(
                    config.lekko_uri,
                    config.owner_name,
                    config.repo_name,
                    MemoryStore(),
                    config.git_repo_path,
                    config.api_key,
                    config.context,
                )
            case CachedServerConfig():
                __client = CachedBackendClient(
                    config.lekko_uri, config.owner_name, config.repo_name, MemoryStore(), config.api_key, config.context
                )
            case _:
                raise exceptions.LekkoError("Unknown client mode")
        return __client


def set_client(client: Client):
    global __client
    with __client_lock:
        if __client:
            __client.close()
        __client = client


def close():
    global __client
    with __client_lock:
        if __client:
            __client.close()
            __client = None


def __get_safe(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with __client_lock:
            if not __client:
                raise exceptions.ClientNotInitialized("lekko_client.initialize() must be called prior to using API")
            return func(*args, **kwargs)

    return wrapper


@__get_safe
def get_bool(namespace: str, key: str, context: Dict[str, Any]) -> bool:
    assert __client
    return __client.get_bool(namespace, key, context)


@__get_safe
def get_int(namespace: str, key: str, context: Dict[str, Any]) -> int:
    assert __client
    return __client.get_int(namespace, key, context)


@__get_safe
def get_float(namespace: str, key: str, context: Dict[str, Any]) -> float:
    assert __client
    return __client.get_float(namespace, key, context)


@__get_safe
def get_string(namespace: str, key: str, context: Dict[str, Any]) -> str:
    assert __client
    return __client.get_string(namespace, key, context)


@__get_safe
def get_json(namespace: str, key: str, context: Dict[str, Any]) -> dict:
    assert __client
    return __client.get_json(namespace, key, context)


@__get_safe
def get_proto(
    namespace: str,
    key: str,
    context: Dict[str, Any],
) -> ProtoMessage:
    assert __client
    return __client.get_proto(namespace, key, context)


@__get_safe
def get_proto_by_type(
    namespace: str,
    key: str,
    context: Dict[str, Any],
    proto_message_type: Type[Client.ProtoType],
) -> Client.ProtoType:
    assert __client
    return __client.get_proto_by_type(namespace, key, context, proto_message_type)



import argparse
import dataclasses
import time
import logging

import grpc
from google.protobuf.json_format import MessageToDict

import lekko_client
from lekko_client.exceptions import LekkoError

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(threadName)s - %(levelname)s - %(message)s')
    parser = argparse.ArgumentParser()
    parser.add_argument("--apikey", type=str)
    parser.add_argument("--mode", type=str, choices=["sidecar", "api", "cachedapi", "cachedgit"])
    parser.add_argument("--owner", type=str)
    parser.add_argument("--repo", type=str)
    parser.add_argument("--namespace", type=str, default="default")
    parser.add_argument("--config", type=str, default="example")
    parser.add_argument(
        "--config-type",
        type=str,
        choices=["bool", "int", "float", "str", "json", "proto"],
        default="bool",
    )
    parser.add_argument("--git-path", type=str, default="")
    parser.add_argument("--proto-type", type=str, default="")
    parser.add_argument("--proto-file", type=str)
    parser.add_argument("-f", action="store_true", help="keep reading in a loop")
    parser.add_argument("-n", type=int, help="number of time to read config in a loop", default=0)
    parser.add_argument("--sleep", type=float, help="time in second to sleep between reading configs", default=0.1)
    args = parser.parse_args()

    base_config = lekko_client.Config(owner_name=args.owner, repo_name=args.repo, api_key=args.apikey)
    if args.mode == "sidecar":
        config = lekko_client.SidecarConfig(**dataclasses.asdict(base_config))
    elif args.mode == "api":
        config = lekko_client.APIConfig(**dataclasses.asdict(base_config))
    elif args.mode == "cachedapi":
        config = lekko_client.CachedServerConfig(**dataclasses.asdict(base_config))
    elif args.mode == "cachedgit":
        config = lekko_client.CachedGitConfig(**dataclasses.asdict(base_config), git_repo_path=args.git_path)
    else:
        raise ValueError("Invalid mode")
    lekko_client.initialize(config)

    count = 0
    start = time.perf_counter_ns()
    while True:
        count += 1
        val = None
        try:
            if args.config_type == "bool":
                val = lekko_client.get_bool(args.namespace, args.config, {})
            elif args.config_type == "int":
                val = lekko_client.get_int(args.namespace, args.config, {})
            elif args.config_type == "str":
                val = lekko_client.get_string(args.namespace, args.config, {})
            elif args.config_type == "json":
                val = lekko_client.get_json(args.namespace, args.config, {})
            elif args.config_type == "float":
                val = lekko_client.get_float(args.namespace, args.config, {})
            elif args.config_type == "proto":
                if not args.proto_file:
                    print(
                        "Must provide a --proto-file. Could be a path or a well-known proto like 'google/protobuf/wrappers.proto'"
                    )
                else:
                    imported_proto = grpc.protos(args.proto_file)

                    if args.proto_type:
                        msg_type = getattr(imported_proto, args.proto_type)
                        val = MessageToDict(lekko_client.get_proto_by_type(args.namespace, args.config, {}, msg_type))
                    else:
                        val = MessageToDict(lekko_client.get_proto(args.namespace, args.config, {}))
            #print(f"Got {val} for config {args.namespace}/{args.config}")
        except LekkoError as e:
            print(f"Failed to get config: {e}")
            if e.__cause__:
                print(f"Caused by: {e.__cause__}")
        time.sleep(args.sleep)
        if not args.f and count > args.n:
            break
    
    total_ms = (time.perf_counter_ns() - start) / 1_000_000
    print(f"total ms: {total_ms}")
    print(f"count: {count}")
    print(f"qps: {count / (total_ms / 1_000)}")
